import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, TypeVar

from typing_extensions import TypeAlias

from whylogs.core.datatypes import AnyType, DataType, Fractional, Integral, String
from whylogs.core.metrics import StandardMetric
from whylogs.core.metrics.metrics import Metric, MetricConfig

logger = logging.getLogger(__name__)


M = TypeVar("M", bound=Metric)
ColumnSchema: TypeAlias = "ColumnSchema"  # type: ignore


class Resolver(ABC):
    """A resolver maps from a column name and a data type to trackers.

    Note that the key of the result dictionaries defines the namespaces of the metrics in the serialized form."""

    @abstractmethod
    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        raise NotImplementedError


class StandardResolver(Resolver):
    """Standard metric resolution with builtin types."""

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        metrics: List[StandardMetric] = [StandardMetric.counts, StandardMetric.types]

        if isinstance(why_type, Integral):
            metrics.append(StandardMetric.distribution)
            metrics.append(StandardMetric.ints)
            metrics.append(StandardMetric.cardinality)
            metrics.append(StandardMetric.frequent_items)
        elif isinstance(why_type, Fractional):
            metrics.append(StandardMetric.cardinality)
            metrics.append(StandardMetric.distribution)
        elif isinstance(why_type, String):  # Catch all category as we map 'object' here
            metrics.append(StandardMetric.cardinality)
            if column_schema.cfg.track_unicode_ranges:
                metrics.append(StandardMetric.unicode_range)
            metrics.append(StandardMetric.distribution)  # 'object' columns can contain Decimal
            metrics.append(StandardMetric.frequent_items)

        if column_schema.cfg.fi_disabled:
            metrics.remove(StandardMetric.frequent_items)

        result: Dict[str, Metric] = {}
        for m in metrics:
            result[m.name] = m.zero(column_schema.cfg)
        return result


class LimitedTrackingResolver(Resolver):
    """Resolver that skips frequent item and cardinality trackers."""

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        metrics: List[StandardMetric] = [StandardMetric.counts, StandardMetric.types]
        if isinstance(why_type, Integral):
            metrics.append(StandardMetric.distribution)
            metrics.append(StandardMetric.ints)

        elif isinstance(why_type, Fractional):
            metrics.append(StandardMetric.distribution)

        result: Dict[str, Metric] = {}
        for m in metrics:
            result[m.name] = m.zero(column_schema.cfg)
        return result


class HistogramCountingTrackingResolver(Resolver):
    """Resolver that only adds distribution tracker."""

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        metrics: List[StandardMetric] = [StandardMetric.distribution]
        result: Dict[str, Metric] = {}
        for m in metrics:
            result[m.name] = m.zero(column_schema.cfg)
        return result


@dataclass
class MetricSpec:
    """
    Specify a Metric to instantiate.
    """

    metric: Any  # Should be a subclass of Metric, it should be the class, not an instance
    config: Optional[MetricConfig] = None  # omit to use default MetricConfig

    def __post_init__(self):
        if not issubclass(self.metric, Metric):
            raise ValueError("MetricSpec: must supply a Metric subclass to MetricSpec")


@dataclass
class ResolverSpec:
    """
    Specify the metrics to instantiate for matching columns. column_name
    takes precedence over column_type. column_type should be a subclass
    of DataType, i.e., AnyType, Frational, Integral, or String. Pass the
    class, not an instance.
    """

    column_name: Optional[str] = None  # TODO: maybe make this a regex
    column_type: Optional[Any] = None
    metrics: List[MetricSpec] = field(default_factory=list)

    def __post_init__(self):
        if self.column_name and self.column_type:
            logger.warning(f"ResolverSpec: column {self.column_name} also specified type, name takes precedence")
        if not (self.column_name or self.column_type):
            raise ValueError("ResolverSpec: resolver specification must supply name or type")

        if self.column_type and not issubclass(self.column_type, DataType):
            raise ValueError("ResolverSpec: resolver specification column type must be a DataType")


# whylabs expects COLUMN_METRICS to be present for every column.
COLUMN_METRICS = [MetricSpec(StandardMetric.counts.value), MetricSpec(StandardMetric.types.value)]


def _allowed_metric(config: MetricConfig, metric: Metric) -> bool:
    """Return False for any metrics turned off in the config"""

    namespace = metric.get_namespace()
    if config.fi_disabled and namespace == "frequent_items":
        return False
    if (not config.track_unicode_ranges) and namespace == "unicode_range":
        return False
    return True


class DeclarativeResolver(Resolver):
    """
    Implements the declarative resolution logic by interpreting a "program"
    of ResolverSpecs
    """

    def add_resolver(self, resolver_spec: ResolverSpec):
        self._resolvers.append(resolver_spec)

    def __init__(self, resolvers: List[ResolverSpec], default_config: Optional[MetricConfig] = None) -> None:
        # Validate resolvers -- must have name xor type, MetricSpec metrics must <: Metric
        self._resolvers = resolvers.copy()

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        result: Dict[str, Metric] = {}
        for resolver_spec in self._resolvers:
            col_name, col_type = resolver_spec.column_name, resolver_spec.column_type
            if (col_name and col_name == name) or (col_name is None and isinstance(why_type, col_type)):  # type: ignore
                for spec in resolver_spec.metrics:
                    config = spec.config or column_schema.cfg
                    if _allowed_metric(config, spec.metric):
                        result[spec.metric.get_namespace()] = spec.metric.zero(config)

        return result


# STANDARD_RESOLVER matches the default DatasetSchema/StandardResolver behavior
STANDARD_RESOLVER = [
    ResolverSpec(
        column_type=Integral,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.ints.value),
            MetricSpec(StandardMetric.cardinality.value),
            MetricSpec(StandardMetric.frequent_items.value),
        ],
    ),
    ResolverSpec(
        column_type=Fractional,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.cardinality.value),
        ],
    ),
    ResolverSpec(
        column_type=String,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.unicode_range.value),
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.cardinality.value),
            MetricSpec(StandardMetric.frequent_items.value),
        ],
    ),
    ResolverSpec(column_type=AnyType, metrics=COLUMN_METRICS),
]

LIMITED_TRACKING_RESOLVER = [
    ResolverSpec(
        column_type=Integral,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.ints.value),
        ],
    ),
    ResolverSpec(
        column_type=Fractional,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
        ],
    ),
    ResolverSpec(column_type=String, metrics=COLUMN_METRICS),
    ResolverSpec(column_type=AnyType, metrics=COLUMN_METRICS),
]


HISTOGRAM_COUNTING_TRACKING_RESOLVER = [
    ResolverSpec(column_type=Integral, metrics=[MetricSpec(StandardMetric.distribution.value)]),
    ResolverSpec(column_type=Fractional, metrics=[MetricSpec(StandardMetric.distribution.value)]),
    ResolverSpec(column_type=String, metrics=[MetricSpec(StandardMetric.distribution.value)]),
    ResolverSpec(column_type=AnyType, metrics=[MetricSpec(StandardMetric.distribution.value)]),
]



from ..metric_constraints import MetricConstraint, MetricsSelector


def condition_meets(column_name: str, condition_name: str) -> MetricConstraint:
    """Checks that all values in column match predicate

    Parameters
    ----------
    column_name : str
        Name of the column to apply the constraint to
    condition_name : str
        Name of the condition that will be applied to each value of the column
    """

    def matches_all(x) -> bool:
        if x.matches[condition_name].value != x.total.value:
            return False
        return True

    constraint = MetricConstraint(
        name=f"{column_name} meets condition {condition_name}",
        condition=matches_all,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="condition_count"),
    )
    return constraint



from typing import Any, Dict, List, Optional

from whylogs.api.logger.logger import Logger
from whylogs.core import DatasetProfile, DatasetSchema
from whylogs.core.stubs import pd


class TransientLogger(Logger):
    def __init__(self, schema: Optional[DatasetSchema] = None):
        super(TransientLogger, self).__init__(schema)

    def _get_matching_profiles(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Dict[str, Any]] = None,
        schema: Optional[DatasetSchema] = None,
    ) -> List[DatasetProfile]:
        active_schema = schema or self._schema
        return [DatasetProfile(schema=active_schema)]



import logging
from typing import Dict, List, Optional

import whylogs.viz.drift.column_drift_algorithms as column_drift_algorithms
from whylogs import DatasetProfileView
from whylogs.viz.enums.enums import PageSpecEnum
from whylogs.viz.extensions.reports.html_report import HTMLReport
from whylogs.viz.utils.html_template_utils import get_compiled_template
from whylogs.viz.utils.profile_viz_calculations import (
    generate_summaries_with_drift_score,
)

logger = logging.getLogger(__name__)


class SummaryDriftReport(HTMLReport):
    _drift_map: Optional[Dict[str, column_drift_algorithms.ColumnDriftAlgorithm]] = None

    def __init__(self, ref_view: DatasetProfileView, target_view: DatasetProfileView, height: Optional[str] = None):
        super().__init__(ref_view=ref_view, target_view=target_view, height=height)

    def report(self) -> str:
        page_spec = PageSpecEnum.SUMMARY_REPORT.value
        template = get_compiled_template(page_spec.html)

        profiles_summary = generate_summaries_with_drift_score(
            self.target_view, self.ref_view, config=None, drift_map=self._drift_map
        )
        rendered_template = template(profiles_summary)
        summary_drift_report = self.display(rendered_template, page_spec)
        return summary_drift_report

    def add_drift_config(
        self, column_names: List[str], algorithm: column_drift_algorithms.ColumnDriftAlgorithm
    ) -> None:
        """Add drift configuration.
        The algorithms and thresholds added through this method will be used to calculate drift scores in the `summary_drift_report()` method.
        If any drift configuration exists, the new configuration will overwrite the standard behavior when appliable.
        If a column has multiple configurations defined, the last one defined will be used.

        Parameters
        ----------
        config: DriftConfig, required
            Drift configuration.

        """
        self._drift_map = {} if not self._drift_map else self._drift_map
        if not isinstance(algorithm, column_drift_algorithms.ColumnDriftAlgorithm):
            raise ValueError("Algorithm must be of class ColumnDriftAlgorithm.")
        if not self.target_view or not self.ref_view:
            logger.error("Set target and reference profiles before adding drift configuration.")
            raise ValueError
        if not algorithm:
            raise ValueError("Drift algorithm cannot be None.")
        if not column_names:
            raise ValueError("Drift configuration must have at least one column name.")
        if column_names:
            for column_name in column_names:
                if column_name not in self.target_view.get_columns().keys():
                    raise ValueError(f"Column {column_name} not found in target profile.")
                if column_name not in self.target_view.get_columns().keys():
                    raise ValueError(f"Column {column_name} not found in reference profile.")
        for column_name in column_names:
            if column_name in self._drift_map:
                logger.warning(f"Overwriting existing drift configuration for column {column_name}.")
            self._drift_map[column_name] = algorithm



import logging
import time
from datetime import datetime
from typing import Dict, Optional

from whylogs.api.logger.result_set import SegmentedResultSet
from whylogs.core.dataset_profile import DatasetProfile
from whylogs.core.schema import DatasetSchema
from whylogs.core.segment import Segment

logger = logging.getLogger(__name__)


class SegmentCache(object):
    """
    Container for segmented profiles.

    Stores DatasetProfiles while processing inputs using a segmentation DatasetSchema.
    The contained profiles may be stored in memory or on disk.
    """

    def __init__(self, schema: DatasetSchema, segments: Optional[Dict[Segment, DatasetProfile]] = None):
        self._schema = schema
        self._cache = segments or dict()

    def get_or_create_matching_profile(self, segment_key: Segment) -> DatasetProfile:
        profile = self._cache.get(segment_key)
        if profile is None:
            profile = DatasetProfile(schema=self._schema)
            self._cache[segment_key] = profile
        return profile

    def get_segments(self) -> Dict[Segment, DatasetProfile]:
        return self._cache

    def flush(self, dataset_timestamp: Optional[datetime]) -> SegmentedResultSet:
        segmented_profiles: Dict[str, Dict[Segment, DatasetProfile]] = dict()
        for segment_key in self._cache:
            segments = segmented_profiles.get(segment_key.parent_id)
            if segments is None:
                segments = dict()
                segmented_profiles[segment_key.parent_id] = segments
            while self._cache[segment_key].is_active:
                time.sleep(1)
            segments[segment_key] = self._cache[segment_key]

        results = SegmentedResultSet(segments=segmented_profiles, partitions=list(self._schema.segments.values()))
        self._cache = dict()
        if dataset_timestamp:
            results.set_dataset_timestamp(dataset_timestamp)
        return results



import datetime
import logging
import os
import tempfile
from typing import IO, Any, Dict, List, Optional, Tuple

import requests  # type: ignore
import whylabs_client  # type: ignore
from whylabs_client import ApiClient, Configuration  # type: ignore
from whylabs_client.api.feature_weights_api import FeatureWeightsApi
from whylabs_client.api.log_api import AsyncLogResponse  # type: ignore
from whylabs_client.api.log_api import (
    LogApi,
    LogAsyncRequest,
    LogReferenceRequest,
    LogReferenceResponse,
)
from whylabs_client.api.models_api import ModelsApi
from whylabs_client.model.column_schema import ColumnSchema
from whylabs_client.rest import ForbiddenException  # type: ignore

from whylogs import __version__ as _version
from whylogs.api.writer import Writer
from whylogs.api.writer.writer import Writable
from whylogs.core import DatasetProfileView
from whylogs.core.dataset_profile import DatasetProfile
from whylogs.core.errors import BadConfigError
from whylogs.core.feature_weights import FeatureWeights
from whylogs.core.utils import deprecated_alias
from whylogs.core.view.segmented_dataset_profile_view import SegmentedDatasetProfileView
from whylogs.migration.uncompound import (
    _uncompound_dataset_profile,
    _uncompound_metric_feature_flag,
)

FIVE_MINUTES_IN_SECONDS = 60 * 5
DAY_IN_SECONDS = 60 * 60 * 24
WEEK_IN_SECONDS = DAY_IN_SECONDS * 7
FIVE_YEARS_IN_SECONDS = DAY_IN_SECONDS * 365 * 5
logger = logging.getLogger(__name__)

API_KEY_ENV = "WHYLABS_API_KEY"


class WhyLabsWriter(Writer):
    """
    A WhyLogs writer to upload DatasetProfileView's onto the WhyLabs platform.

    >**IMPORTANT**: In order to correctly send your profiles over, make sure you have
    the following environment variables set: `[WHYLABS_ORG_ID, WHYLABS_API_KEY, WHYLABS_DEFAULT_DATASET_ID]`. You
    can also set them with the option method or within the constructor, although it
    is highly recommended you don't persist credentials in code!

    Parameters
    ----------
    org_id: str, optional
        Your organization ID on the WhyLabs platform. If not informed, will get the mentioned environment
        variable instead.

    api_key: str, optional
        The API Key to authenticate with WhyLabs. If not informed, will get the mentioned environment
        variable instead.

    dataset_id: str, optional
        The Dataset ID, often referred to as model ID, that will tie the profile to your tracked data.
        If not informed, will get the mentioned environment variable instead.
    Returns
    -------
        None

    Examples
    --------
    An example usage of this method can be represented with the simple code above. Here we
    are assuming the user has already assigned a pandas.DataFrame to the `df` variable.

    ```python
    import whylogs as why

    profile = why.log(pandas=df)
    profile.writer("whylabs").write()
    ```

    """

    def __init__(self, org_id: Optional[str] = None, api_key: Optional[str] = None, dataset_id: Optional[str] = None):
        self._org_id = org_id or os.environ.get("WHYLABS_DEFAULT_ORG_ID")
        self._api_key = api_key or os.environ.get("WHYLABS_API_KEY")
        self._api_key_id = ""
        self._dataset_id = dataset_id or os.environ.get("WHYLABS_DEFAULT_DATASET_ID")
        self._feature_weights = None
        self.whylabs_api_endpoint = os.environ.get("WHYLABS_API_ENDPOINT") or "https://api.whylabsapp.com"
        self._reference_profile_name = os.environ.get("WHYLABS_REFERENCE_PROFILE_NAME")
        self._api_log_client: Optional[ApiClient] = None
        self._ssl_ca_cert: Optional[str] = None
        self._configuration: Optional[Configuration] = None

    def check_interval(self, interval_seconds: int) -> None:
        if interval_seconds < FIVE_MINUTES_IN_SECONDS:
            raise BadConfigError("Bad WhyLabsWriter config: interval must be greater or equal to five minutes")

    def option(  # type: ignore
        self,
        org_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        api_key: Optional[str] = None,
        reference_profile_name: Optional[str] = None,
        configuration: Optional[Configuration] = None,
        ssl_ca_cert: Optional[str] = None,
    ) -> "WhyLabsWriter":
        """

        Parameters
        ----------
        org_id the organization ID
        dataset_id the dataset Id
        api_key the API key
        reference_profile_name the name of the reference profile
        configuration the additional configuration for the REST client

        Returns a "WhyLabsWriter" with these options configured
        -------

        """
        if dataset_id is not None:
            self._dataset_id = dataset_id
        if org_id is not None:
            self._org_id = org_id
        if api_key is not None:
            self._api_key = api_key
        if reference_profile_name is not None:
            self._reference_profile_name = reference_profile_name
        if configuration is not None:
            self._configuration = configuration
        if ssl_ca_cert is not None:
            self._ssl_ca_cert = ssl_ca_cert
        return self

    def _tag_columns(self, columns: List[str], value: str) -> Tuple[bool, str]:
        """Sets the column as an input or output for the specified dataset.

        Parameters
        ----------
        columns : List[str]
            The list of column names you want to tag as input or output

        value: str
            For example we can pass in "output" to configure the column as an output.

        Returns
        -------
        Tuple[bool, str]
            Tuple with a boolean indicating success or failure: e.g. (True, "column prediction was updated to output") and string with status message.
        """
        self._validate_org_and_dataset()
        self._validate_api_key()
        results: Dict[str, str] = dict()
        all_sucessful = True
        # TODO: update the list of columns at once, support arbitrary tags as well.
        for column_name in columns:
            status = self._put_column_schema(column_name, value)
            results[column_name] = status[1]
            if status[0] != 200:
                all_sucessful = False
        return all_sucessful, str(results)

    def tag_output_columns(self, columns: List[str]) -> Tuple[bool, str]:
        """Sets the list of columns as output columns for the specified dataset.

        Parameters
        ----------
        columns : List[str]
            The list of column names you want to tag as outputs.

        Returns
        -------
        Tuple[bool, str]
            Tuple with a boolean indicating success or failure: e.g. (True, "column prediction was updated to output") and string with status message.
        """

        return self._tag_columns(columns, "output")

    def tag_input_columns(self, columns: List[str]) -> Tuple[bool, str]:
        """Sets the list of columns as input columns for the specified dataset.

        Parameters
        ----------
        columns : List[str]
            The list of column names you want to tag as inputs.

        Returns
        -------
        Tuple[bool, str]
            Tuple with a boolean indicating success or failure: e.g. (True, "column [output_voltage] updated to input") and string with status message.
        """

        return self._tag_columns(columns, "input")

    def write_feature_weights(self, file: FeatureWeights, **kwargs: Any) -> Tuple[bool, str]:
        """Put feature weights for the specified dataset.

        Parameters
        ----------
        file : FeatureWeights
            FeatureWeights object representing the Feature Weights for the specified dataset

        Returns
        -------
        Tuple[bool, str]
            Tuple with a boolean (1-success, 0-fail) and string with the request's status code.
        """
        self._feature_weights = file.to_dict()
        if kwargs.get("dataset_id") is not None:
            self._dataset_id = kwargs.get("dataset_id")
        return self._do_upload_feature_weights()

    def get_feature_weights(self, **kwargs: Any) -> Optional[FeatureWeights]:
        """Get latest version for the feature weights for the specified dataset

        Returns
        -------
        FeatureWeightResponse
            Response of the GET request, with segmentWeights and metadata.
        """
        if kwargs.get("dataset_id") is not None:
            self._dataset_id = kwargs.get("dataset_id")

        result = self._do_get_feature_weights()
        feature_weights_set = result.get("segmentWeights")
        metadata = result.get("metadata")
        if feature_weights_set and isinstance(feature_weights_set, list):
            feature_weights = FeatureWeights(weights=feature_weights_set[0]["weights"], metadata=metadata)
            return feature_weights
        return None

    @deprecated_alias(profile="file")
    def write(self, file: Writable, **kwargs: Any) -> Tuple[bool, str]:
        if isinstance(file, FeatureWeights):
            return self.write_feature_weights(file, **kwargs)

        view = file.view() if isinstance(file, DatasetProfile) else file
        has_segments = isinstance(view, SegmentedDatasetProfileView)

        if not has_segments and not isinstance(view, DatasetProfileView):
            raise ValueError(
                "You must pass either a DatasetProfile or a DatasetProfileView in order to use this writer!"
            )

        if _uncompound_metric_feature_flag():
            if has_segments:
                updated_profile_view = _uncompound_dataset_profile(view.profile_view)
                view = SegmentedDatasetProfileView(
                    profile_view=updated_profile_view, segment=view._segment, partition=view._partition
                )

            else:
                view = _uncompound_dataset_profile(view)

        if kwargs.get("dataset_id") is not None:
            self._dataset_id = kwargs.get("dataset_id")

        with tempfile.NamedTemporaryFile() as tmp_file:
            if has_segments:
                view.write(file=tmp_file, use_v0=True)
            else:
                view.write(file=tmp_file)
            tmp_file.flush()
            tmp_file.seek(0)
            utc_now = datetime.datetime.now(datetime.timezone.utc)
            dataset_timestamp = view.dataset_timestamp or utc_now
            stamp = dataset_timestamp.timestamp()
            time_delta_seconds = utc_now.timestamp() - stamp
            if time_delta_seconds < 0:
                logger.warning(
                    f"About to upload a profile with a dataset_timestamp that is in the future: {time_delta_seconds}s old."
                )
            elif time_delta_seconds > WEEK_IN_SECONDS:
                if time_delta_seconds > FIVE_YEARS_IN_SECONDS:
                    logger.error(
                        f"A profile being uploaded to WhyLabs has a dataset_timestamp of({dataset_timestamp}) "
                        f"compared to current datetime: {utc_now}. Uploads of profiles older than 5 years "
                        "might not be monitored in WhyLabs and may take up to 24 hours to show up."
                    )
                else:
                    logger.warning(
                        f"A profile being uploaded to WhyLabs has a dataset_timestamp of {dataset_timestamp} "
                        f"which is older than 7 days compared to {utc_now}. These profiles should be processed within 24 hours."
                    )

            if stamp <= 0:
                logger.error(
                    f"Profiles should have timestamps greater than 0, but found a timestamp of {stamp}"
                    f" and current timestamp is {utc_now.timestamp()}, this is likely an error."
                )

            dataset_timestamp_epoch = int(stamp * 1000)
            response = self._do_upload(
                dataset_timestamp=dataset_timestamp_epoch,
                profile_file=tmp_file,
            )
        # TODO: retry
        return response

    def _validate_api_key(self) -> None:
        if self._api_key is None:
            raise ValueError("Missing API key. Set it via WHYLABS_API_KEY environment variable or as an api_key option")
        if len(self._api_key) < 12:
            raise ValueError("API key too short")
        if len(self._api_key) > 64:
            raise ValueError("API key too long")
        if self._api_key[10] != ".":
            raise ValueError("Invalid format. Expecting a dot at an index 10")
        self._api_key_id = self._api_key[:10]

    def _validate_org_and_dataset(self) -> None:
        if self._org_id is None:
            raise EnvironmentError(
                "Missing organization ID. Specify it via option or WHYLABS_DEFAULT_ORG_ID " "environment variable"
            )
        if self._dataset_id is None:
            raise EnvironmentError(
                "Missing dataset ID. Specify it via WHYLABS_DEFAULT_DATASET_ID environment "
                "variable or on your write method"
            )

    def _do_get_feature_weights(self):
        """Get latest version for the feature weights for the specified dataset

        Returns
        -------
            Response of the GET request, with segmentWeights and metadata.
        """
        self._validate_org_and_dataset()
        self._validate_api_key()

        result = self._get_column_weights()
        return result

    def _do_upload_feature_weights(self) -> Tuple[bool, str]:
        """Put feature weights for the specified dataset.

        Returns
        -------
        Tuple[bool, str]
            Tuple with a boolean (1-success, 0-fail) and string with the request's status code.
        """

        self._validate_org_and_dataset()
        self._validate_api_key()

        result = self._put_feature_weights()
        if result == 200:
            return True, str(result)
        else:
            return False, str(result)

    def _copy_file(self, profile_file: IO[bytes], upload_url: str, dataset_timestamp: int, api_key_id: Optional[str]):
        http_response = requests.put(upload_url, data=profile_file.read())
        is_successful = False
        if http_response.status_code == 200:
            is_successful = True
            logger.info(
                f"Done uploading {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                f"{self.whylabs_api_endpoint} with API token ID: {api_key_id}"
            )
        return is_successful, http_response.text

    def _do_upload(
        self, dataset_timestamp: int, profile_path: Optional[str] = None, profile_file: Optional[IO[bytes]] = None
    ) -> Tuple[bool, str]:
        assert profile_path or profile_file, "Either a file or file path must be specified when uploading profiles"
        self._validate_org_and_dataset()
        self._validate_api_key()

        logger.debug("Generating the upload URL")
        upload_url = self._get_upload_url(dataset_timestamp=dataset_timestamp)
        api_key_id = self._api_key[:10] if self._api_key else None
        try:
            if profile_file:
                return self._copy_file(profile_file, upload_url, dataset_timestamp, api_key_id)
            elif profile_path:
                with open(profile_path, "rb") as f:
                    return self._copy_file(f, upload_url, dataset_timestamp, api_key_id)
        except requests.RequestException as e:
            logger.info(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                + f"{self.whylabs_api_endpoint} with API token ID: {api_key_id}. Error occurred: {e}"
            )
            return False, str(e)
        return False, "Either a profile_file or profile_path must be specified when uploading profiles to WhyLabs!"

    def _get_or_create_feature_weights_client(self) -> FeatureWeightsApi:
        environment_api_key = os.environ.get(API_KEY_ENV)

        if self._api_log_client is None:
            self._refresh_api_client()
        elif environment_api_key is not None and self._api_key != environment_api_key:
            logger.warning(f"Updating API key ID from: {self._api_key_id} to: {environment_api_key[:10]}")
            self._api_key = environment_api_key
            self._refresh_api_client()

        return FeatureWeightsApi(self._api_log_client)

    def _get_or_create_models_client(self) -> ModelsApi:
        environment_api_key = os.environ.get(API_KEY_ENV)

        if self._api_log_client is None:
            self._refresh_api_client()
        elif environment_api_key is not None and self._api_key != environment_api_key:
            logger.warning(f"Updating API key ID from: {self._api_key_id} to: {environment_api_key[:10]}")
            self._api_key = environment_api_key
            self._refresh_api_client()

        return ModelsApi(self._api_log_client)

    def _get_or_create_api_log_client(self) -> LogApi:
        environment_api_key = os.environ.get(API_KEY_ENV)

        if self._api_log_client is None:
            self._refresh_api_client()
        elif environment_api_key is not None and self._api_key != environment_api_key:
            logger.warning(f"Updating API key ID from: {self._api_key_id} to: {environment_api_key[:10]}")
            self._api_key = environment_api_key
            self._refresh_api_client()

        return LogApi(self._api_log_client)

    def _refresh_api_client(self) -> None:
        self._validate_api_key()

        if self._configuration is None:
            config = whylabs_client.Configuration()
        else:
            config = self._configuration
        config.ssl_ca_cert = self._ssl_ca_cert
        config.host = self.whylabs_api_endpoint
        config.api_key = {"ApiKeyAuth": self._api_key}
        config.discard_unknown_keys = True

        client = whylabs_client.ApiClient(config)
        client.user_agent = f"whylogs/python/{_version}"
        self._api_log_client = client

    def _get_rest_config(self) -> Configuration:
        if self._configuration is None:
            config = whylabs_client.Configuration()
        else:
            config = self._configuration

        config.host = self.whylabs_api_endpoint
        config.api_key = {"ApiKeyAuth": self._api_key}
        config.discard_unknown_keys = True
        return config

    @staticmethod
    def _build_log_async_request(dataset_timestamp: int) -> LogAsyncRequest:
        return LogAsyncRequest(dataset_timestamp=dataset_timestamp, segment_tags=[])

    @staticmethod
    def _build_log_reference_request(dataset_timestamp: int, alias: Optional[str] = None) -> LogReferenceRequest:
        return LogReferenceRequest(dataset_timestamp=dataset_timestamp, alias=alias)

    def _get_column_weights(self):
        feature_weight_api = self._get_or_create_feature_weights_client()
        try:
            result = feature_weight_api.get_column_weights(
                org_id=self._org_id,
                dataset_id=self._dataset_id,
            )
            return result
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id} to "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e

    def _put_feature_weights(self):
        feature_weight_api = self._get_or_create_feature_weights_client()
        try:
            result = feature_weight_api.put_column_weights(
                org_id=self._org_id,
                dataset_id=self._dataset_id,
                body={
                    "segmentWeights": [self._feature_weights],
                },
                _return_http_data_only=False,
            )
            return result[1]
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id} to "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e

    def _get_existing_column_schema(self, model_api_instance, column_name) -> Optional[ColumnSchema]:
        try:
            # TODO: remove when whylabs supports merge writes.
            existing_schema = model_api_instance.get_entity_schema_column(
                org_id=self._org_id, dataset_id=self._dataset_id, column_id=column_name
            )

            if existing_schema:
                data_type = existing_schema.get("dataType")
                discreteness = existing_schema.get("discreteness")
                existing_classification = existing_schema.get("classifier")
                column_schema = ColumnSchema(
                    classifier=existing_classification, data_type=data_type, discreteness=discreteness
                )
                return column_schema
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to set column outputs {self._org_id}/{self._dataset_id} for column name: ({column_name}) "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e
        return None

    def _column_schema_needs_update(self, column_schema: ColumnSchema, new_classification: str) -> bool:
        existing_classification = column_schema.classifier
        if not existing_classification:
            return True
        return existing_classification != new_classification

    def _put_column_schema(self, column_name: str, value: str) -> Tuple[int, str]:
        model_api_instance = self._get_or_create_models_client()

        # TODO: simplify after ColumnSchema updates support merge writes, can remove the read here.
        column_schema: ColumnSchema = self._get_existing_column_schema(model_api_instance, column_name)
        if self._column_schema_needs_update(column_schema, value):
            updated_column_schema = ColumnSchema(
                classifier=value, data_type=column_schema.data_type, discreteness=column_schema.discreteness
            )
            try:
                # TODO: remove when whylabs supports merge writes.
                model_api_instance.put_entity_schema_column(
                    self._org_id, self._dataset_id, column_name, updated_column_schema
                )
                return (
                    200,
                    f"{column_name} updated from {column_schema.classifier} to {updated_column_schema.classifier}",
                )
            except ForbiddenException as e:
                api_key_id = self._api_key[:10] if self._api_key else None
                logger.exception(
                    f"Failed to set column outputs {self._org_id}/{self._dataset_id} for column name: ({column_name}) "
                    f"{self.whylabs_api_endpoint}"
                    f" with API token ID: {api_key_id}"
                )
                raise e
        else:
            no_update_made_message = f"column {column_name} was already classified {column_schema.classifier}."
            logger.info(no_update_made_message)
            return (200, no_update_made_message)

    def _post_log_async(self, request: LogAsyncRequest, dataset_timestamp: int) -> AsyncLogResponse:
        log_api = self._get_or_create_api_log_client()
        try:
            result = log_api.log_async(org_id=self._org_id, dataset_id=self._dataset_id, log_async_request=request)
            return result
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e

    def _post_log_reference(self, request: LogAsyncRequest, dataset_timestamp: int) -> LogReferenceResponse:
        log_api = self._get_or_create_api_log_client()
        try:
            async_result = log_api.log_reference(
                org_id=self._org_id, model_id=self._dataset_id, log_reference_request=request, async_req=True
            )
            result = async_result.get()
            return result
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e

    def _get_upload_url(self, dataset_timestamp: int) -> str:
        if self._reference_profile_name is not None:
            request = self._build_log_reference_request(dataset_timestamp, alias=self._reference_profile_name)
            res = self._post_log_reference(request=request, dataset_timestamp=dataset_timestamp)
            upload_url = res["upload_url"]
        else:
            request = self._build_log_async_request(dataset_timestamp)
            res = self._post_log_async(request=request, dataset_timestamp=dataset_timestamp)
            upload_url = res["upload_url"]
        return upload_url



import json
from typing import Any, Dict, Tuple, Union

from whylogs.api.writer import Writers
from whylogs.api.writer.writer import Optional, Writable, Writer
from whylogs.core import Segment


class FeatureWeights(Writable):
    def __init__(self, weights: Dict[str, float], segment: Optional[Segment] = None, metadata: Optional[Dict] = None):
        """Feature Weights

        Parameters
        ----------
        weights : Dict[str, float]
            Feature weights
        segment : str, optional
            If segment is None, weights are considered to be for the complete unsegmented data, by default None
        """
        self.weights = weights
        if segment:
            raise NotImplementedError("Segmented Feature Weights is currently not supported.")
        self.segment = segment
        self.metadata = metadata

    def writer(self, name: str = "local") -> "FeatureWeightWriter":
        writer = Writers.get(name)
        return FeatureWeightWriter(feature_weight=self, writer=writer)

    def get_default_path(self) -> str:
        return f"feature_weights_{self.segment}.json"

    def write(self, path: Optional[str] = None, **kwargs: Any) -> Tuple[bool, str]:
        return (False, "write() not implemented on feature weights, use with WhyLabsWriter.")

    def to_json(self) -> str:
        return json.dumps({"segment": self.segment, "weights": self.weights})

    def to_dict(self) -> Dict[str, Union[Optional[Segment], Optional[float]]]:
        return {"segment": self.segment, "weights": self.weights}


class FeatureWeightWriter(Writer):
    def __init__(self, feature_weight, writer: Writer) -> None:
        self._feature_weight = feature_weight
        self._writer = writer

    def option(self, **kwargs) -> "FeatureWeightWriter":
        self._writer.option(**kwargs)
        return self

    def write(self, **kwargs: Any) -> Tuple[bool, str]:
        """Put feature weights for the specified dataset.

        Returns
        -------
        Tuple[bool, str]
            Tuple with a boolean (1-success, 0-fail) and string with the PUT request's status code.
        """
        return self._writer.write(file=self._feature_weight, **kwargs)

    def get_feature_weights(self, **kwargs: Any) -> Optional[FeatureWeights]:
        """Get latest version for the feature weights for the specified dataset

        Returns
        -------
        FeatureWeightResponse
            Response of the GET request, with segmentWeights and metadata.
        """
        return self._writer.get_feature_weights(**kwargs)



from typing import Union

from whylogs.core.configs import SummaryConfig
from whylogs.core.dataset_profile import DatasetProfile
from whylogs.core.metrics.metrics import Metric
from whylogs.core.relations import ValueGetter
from whylogs.core.view.dataset_profile_view import DatasetProfileView


class MetricGetter(ValueGetter):
    def __init__(self, metric: Metric, path: str) -> None:
        self._metric = metric
        self._path = path
        summary = self._metric.to_summary_dict(SummaryConfig())
        if path not in summary:
            raise ValueError(f"{path} is not available in {metric.namespace}")

    def __call__(self) -> Union[str, int, float]:
        summary = self._metric.to_summary_dict(SummaryConfig())
        return summary[self._path]

    def serialize(self) -> str:
        return f"::{self._metric.namespace}/{self._path}"


class ProfileGetter(ValueGetter):
    def __init__(self, profile: Union[DatasetProfile, DatasetProfileView], column_name: str, path: str) -> None:
        self._profile = profile
        self._column_name = column_name
        self._path = path

    def __call__(self) -> Union[str, int, float]:
        view = self._profile if isinstance(self._profile, DatasetProfileView) else self._profile.view()
        col_prof = view.get_column(self._column_name)
        if not col_prof:
            raise ValueError(f"Column {self._column_name} not found in profile")
        summary = col_prof.to_summary_dict()
        if self._path not in summary:
            raise ValueError(f"{self._path} is not available in {self._column_name} profile")
        return summary[self._path]

    def serialize(self) -> str:
        return f":{self._column_name}:{self._path}"



import logging
import os
import sqlite3
from datetime import timedelta, timezone

from whylogs.api.store import ProfileStore
from whylogs.api.store.query import BaseQuery, DatasetIdQuery, DateQuery
from whylogs.core import DatasetProfile, DatasetProfileView

logger = logging.getLogger(__name__)


class SQLiteStore(ProfileStore):
    def __init__(self):
        self._db_location = os.getenv("SQLITE_STORE_LOCATION")
        if self._db_location is None:
            raise ValueError("You must define a store with the SQLITE_STORE_LOCATION env var")
        self.__conn = sqlite3.connect(database=self._db_location, check_same_thread=False)
        self.__cur = self.__conn.cursor()
        self._init_db()

    def __enter__(self) -> "SQLiteStore":
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        try:
            self.__cur.close()
            self.__conn.close()
        except exception_type:
            logger.error(traceback)
            raise exception_value

    def __del__(self) -> None:
        try:
            self.__cur.close()
            self.__conn.close()
        except Exception as e:
            logger.debug(f"Connection not established. Error: {e}")

    def _init_db(self):
        init_db_query = "CREATE TABLE IF NOT EXISTS profile_store(id string, date datetime, profile blob);"
        self.__cur.execute(init_db_query)

    def list(self):
        sql_query = "SELECT DISTINCT id FROM profile_store;"
        response = self.__cur.execute(sql_query)
        return [item[0] for item in response]

    def get(self, query: BaseQuery) -> DatasetProfileView:
        available_profiles = self.list()
        if query.dataset_id not in available_profiles:
            return DatasetProfile().view()
        if isinstance(query, DatasetIdQuery):
            sql_query = "SELECT profile FROM profile_store WHERE id = ?"
            data_tuple = (query.dataset_id,)
            response = self.__cur.execute(sql_query, data_tuple).fetchall()
        elif isinstance(query, DateQuery):
            start_date = query.start_date.replace(minute=0).astimezone(tz=timezone.utc)
            end_date = query.end_date.astimezone(tz=timezone.utc) + timedelta(minutes=1)
            sql_query = f"SELECT profile FROM profile_store WHERE id = '{query.dataset_id}' AND date BETWEEN '{start_date}' AND '{end_date}';"
            response = self.__cur.execute(sql_query).fetchall()
        else:
            logger.error("Define a supported Query object")
            raise ValueError

        profile_view = DatasetProfileView.zero()
        for item in response:
            profile_view = profile_view.merge(DatasetProfileView.deserialize(item[0]))
        return profile_view

    def _insert_blob(self, profile_view: DatasetProfileView, dataset_id: str):
        try:
            profile_date = profile_view.creation_timestamp
            serialized_profile = profile_view.serialize()
            query = "INSERT INTO profile_store (id, date, profile) VALUES (?, ?, ?);"
            values_tuple = (dataset_id, profile_date, serialized_profile)
            self.__cur.execute(query, values_tuple)
            self.__conn.commit()
            logger.debug(f"Profile {dataset_id} successfully written to the Store!")
        except sqlite3.Error as e:
            logger.error(f"Error: {e}")

    @staticmethod
    def _get_merge_period() -> int:
        merge_period = os.getenv("MERGE_PROFILE_PERIOD_HOURS")
        if not merge_period:
            raise ValueError("You must define a MERGE_PROFILE_PERIOD_HOURS env var")
        return int(merge_period)

    def _check_if_profile_exists(self, profile_view: DatasetProfileView, dataset_id: str) -> DatasetProfileView:
        merge_period = self._get_merge_period()

        start_date = profile_view.creation_timestamp - timedelta(hours=merge_period)
        end_date = profile_view.creation_timestamp + timedelta(minutes=1)
        date_query = DateQuery(dataset_id=dataset_id, start_date=start_date, end_date=end_date)

        existing_profile = self.get(query=date_query)
        return existing_profile

    def _update_profile(self, profile_view: DatasetProfileView, dataset_id: str) -> None:
        merge_period = self._get_merge_period()

        serialized_profile = profile_view.serialize()
        start_date = profile_view.creation_timestamp - timedelta(hours=merge_period)
        end_date = profile_view.creation_timestamp + timedelta(minutes=1)

        update_values = (serialized_profile, dataset_id, start_date, end_date)
        update_query = """
            UPDATE profile_store
            SET profile = ?
            WHERE id = ?
            AND date BETWEEN ? AND ?
        """
        try:
            self.__cur.execute(update_query, update_values)
        except sqlite3.Error as e:
            logger.error(f"Error: {e}")

    def write(self, profile_view: DatasetProfileView, dataset_id: str):
        existing_profile = self._check_if_profile_exists(profile_view=profile_view, dataset_id=dataset_id)
        if existing_profile.get_columns() == {}:
            self._insert_blob(profile_view=profile_view, dataset_id=dataset_id)
        else:
            merged_profile = existing_profile.merge(profile_view)
            self._update_profile(profile_view=merged_profile, dataset_id=dataset_id)



import logging
from typing import Dict

from whylogs.core import ColumnProfileView, DatasetProfileView
from whylogs.core.metrics import Metric
from whylogs.core.metrics.compound_metric import CompoundMetric
from whylogs.core.metrics.multimetric import MultiMetric

try:
    from PIL.Image import Image as ImageType  # noqa trigger ImportError if unavailable

    from whylogs.extras.image_metric import ImageMetric
except ImportError:
    # dummy ImageMetric class so nothing will be an instance of it
    class ImageMetric:  # type: ignore[no-redef]
        pass


logger = logging.getLogger(__name__)


def _uncompound_metric_feature_flag() -> bool:
    """
    v0 whylabs doesn't understand compound metrics. If this is True, turn
    each submetric in a compound metric into its own column in the profile
    so that v0 whylabs will only see metrics it understands.
    """
    return True


def _v0_compatible_image_feature_flag() -> bool:
    """
    v0 whylogs only supported logging a single image in a profile.
    v1 supports multiple image in a profile by giving each image "channel"
    a unique name. If this returns True, use the old V0 column naming
    convention that only supports a single image.
    """
    return False


def _uncompounded_column_name(column_name: str, metric_name: str, submetric_name: str, metric: Metric) -> str:
    if isinstance(metric, ImageMetric) and _v0_compatible_image_feature_flag():
        return submetric_name
    return f"{column_name}.{metric_name}.{submetric_name}"


def _uncompound_metric(col_name: str, metric_name: str, metric: CompoundMetric) -> Dict[str, ColumnProfileView]:
    result: Dict[str, ColumnProfileView] = dict()
    for submetric_name, submetric in metric.submetrics.items():
        new_col_name = _uncompounded_column_name(col_name, metric_name, submetric_name, metric)
        result[new_col_name] = ColumnProfileView({submetric.namespace: submetric})
    return result


def _uncompounded_multimetric_column_name(column_name: str, submetric_name: str, metric: Metric) -> str:
    if isinstance(metric, ImageMetric) and _v0_compatible_image_feature_flag():
        return submetric_name
    return f"{column_name}.{submetric_name}"


def _uncompound_multimetric(col_name: str, metric_name: str, metric: MultiMetric) -> Dict[str, ColumnProfileView]:
    result: Dict[str, ColumnProfileView] = dict()
    for submetric_name, submetrics in metric.submetrics.items():
        new_col_name = _uncompounded_multimetric_column_name(col_name, submetric_name, metric)
        result[new_col_name] = ColumnProfileView(submetrics)

    return result


def _uncompound_dataset_profile(prof: DatasetProfileView) -> DatasetProfileView:
    """
    v0 whylabs doesn't understand compound metrics. This creates a new column for
    each submetric in a compound metric so that whylabs only sees metrics it understands.
    """
    new_prof = DatasetProfileView(
        columns=prof._columns,
        dataset_timestamp=prof._dataset_timestamp,
        creation_timestamp=prof._creation_timestamp,
        metrics=prof._metrics,
    )
    new_columns: Dict[str, ColumnProfileView] = dict()
    for col_name, col_prof in new_prof._columns.items():
        for metric_name, metric in col_prof._metrics.items():
            if isinstance(metric, CompoundMetric):
                new_columns.update(_uncompound_metric(col_name, metric_name, metric))
            if isinstance(metric, MultiMetric):
                new_columns.update(_uncompound_multimetric(col_name, metric_name, metric))

    new_prof._columns.update(new_columns)
    return new_prof



from collections import namedtuple
from copy import deepcopy
from dataclasses import dataclass
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional, Tuple

from whylogs.core.metrics.metrics import Metric
from whylogs.core.utils import deprecated
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.core.view.dataset_profile_view import DatasetProfileView

logger = getLogger(__name__)


ReportResult = namedtuple("ReportResult", "name passed failed summary")


@dataclass
class MetricsSelector:
    metric_name: str
    column_name: Optional[str] = None  # <- TODO: return a collection of columns
    metrics_resolver: Optional[Callable[[DatasetProfileView], List[Metric]]] = None

    def column_profile(self, profile: DatasetProfileView) -> Optional[ColumnProfileView]:
        if profile is None:
            return None
        column_profile_view = profile.get_column(self.column_name)
        if column_profile_view is None:
            logger.warning(
                f"No column name {self.column_name} found in profile, available columns are: {profile.get_columns()}"
            )
        return column_profile_view

    def apply(self, profile: DatasetProfileView) -> List[Metric]:
        if self.metrics_resolver is not None:
            custom_result = self.metrics_resolver(profile)
            if isinstance(custom_result, List):
                return custom_result
            else:
                raise ValueError("metrics_resolver must return a list of Metrics if defined!")
        results: List[Metric] = []
        column_profile_view = self.column_profile(profile)
        if column_profile_view is None:
            return results
        metric = column_profile_view.get_metric(self.metric_name)
        if metric is None:
            logger.warning(
                f"{self.metric_name} not found in {self.column_name} available metrics are: "
                f"{column_profile_view.get_metric_component_paths()}"
            )
        else:
            results.append(metric)
        return results


def pretty_display(metric_name: str, summary: Dict[str, Any]) -> Dict[str, Any]:
    pretty_summary: Dict[str, Any] = {"metric": metric_name}

    for key, item in summary.items():
        if key != "frequent_strings":
            pretty_summary[key] = item
        else:
            pretty_summary[f"{key}_top_10"] = [f"{x.value}:{x.est}" for x in item[:10]]
    return pretty_summary


@dataclass
class MetricConstraint:
    condition: Callable[[Metric], bool]
    name: str
    metric_selector: MetricsSelector
    require_column_existence: bool = True

    def _validate_metrics(self, metrics: List[Metric]) -> bool:
        for metric in metrics:
            if not self.condition(metric):
                return False
        return True

    def _get_metric_summary(self, metrics: List[Metric]) -> Optional[Dict[str, Any]]:
        if len(metrics) == 1:  # Only returns a summary for single metrics.
            metric_summary = pretty_display(metrics[0].namespace, metrics[0].to_summary_dict())
            return metric_summary
        return None

    @deprecated(message="Please use validate_profile()")
    def validate(self, dataset_profile: DatasetProfileView) -> bool:
        # custom metric resolver allows empty metrics
        if self.metric_selector is None:
            raise ValueError("can't call validate with an empty metric selector")
        metric_selector: MetricsSelector = self.metric_selector
        if metric_selector.metrics_resolver is not None:
            metrics = metric_selector.apply(dataset_profile)
            logger.info(f"validate using custom metric selector and found {metrics}")
            return self._validate_metrics(metrics)

        # standard selector requires a column be resolved
        column_profile = metric_selector.column_profile(dataset_profile)
        if column_profile is None:
            if self.require_column_existence:
                raise ValueError(
                    f"Could not get column {metric_selector.column_name}" f"from column {dataset_profile.get_columns()}"
                )
            else:
                logger.info(
                    f"validate could not find column {metric_selector.column_name} "
                    "but require_column_existence is false so returning True"
                )
                return True
        metrics = metric_selector.apply(dataset_profile)
        if metrics is None or len(metrics) == 0:
            if self.require_column_existence:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} so returning False. Available metrics on column are: "
                    f"{column_profile.get_metric_component_paths()}"
                )
                return False
            else:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} but require_column_existence is set so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return True
        return self._validate_metrics(metrics)

    def validate_profile(self, dataset_profile: DatasetProfileView) -> Tuple[bool, Optional[Dict[str, Any]]]:
        # custom metric resolver allows empty metrics
        if self.metric_selector is None:
            raise ValueError("can't call validate with an empty metric selector")
        metric_selector: MetricsSelector = self.metric_selector
        if metric_selector.metrics_resolver is not None:
            metrics = metric_selector.apply(dataset_profile)
            logger.info(f"validate using custom metric selector and found {metrics}")
            validate_result = self._validate_metrics(metrics)
            metric_summary = self._get_metric_summary(metrics)
            return (validate_result, metric_summary)

        # standard selector requires a column be resolved
        column_profile = metric_selector.column_profile(dataset_profile)
        if column_profile is None:
            if self.require_column_existence:
                raise ValueError(
                    f"Could not get column {metric_selector.column_name}" f"from column {dataset_profile.get_columns()}"
                )
            else:
                logger.info(
                    f"validate could not find column {metric_selector.column_name} "
                    "but require_column_existence is false so returning True"
                )
                return (True, None)
        metrics = metric_selector.apply(dataset_profile)
        if metrics is None or len(metrics) == 0:
            if self.require_column_existence:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} so returning False. Available metrics on column are: "
                    f"{column_profile.get_metric_component_paths()}"
                )
                return (False, None)
            else:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} but require_column_existence is false so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return (True, None)
        if len(metrics) == 1 and metrics[0].namespace == "distribution":
            if metrics[0].kll.value.is_empty() and self.require_column_existence:
                logger.info(
                    f"validate reached empty distribution metrics from column "
                    f"{metric_selector.column_name} but require_column_existence is set so returning False. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return (False, None)
            if metrics[0].kll.value.is_empty() and not self.require_column_existence:
                logger.info(
                    f"validate reached empty distribution metrics from column "
                    f"{metric_selector.column_name} but require_column_existence is false so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return (True, None)

        validate_result = self._validate_metrics(metrics)
        metric_summary = self._get_metric_summary(metrics)
        return (validate_result, metric_summary)


class Constraints:
    column_constraints: Dict[str, Dict[str, MetricConstraint]]
    dataset_constraints: Dict[str, MetricConstraint]
    dataset_profile_view: Optional[DatasetProfileView]

    def __init__(
        self,
        dataset_profile_view: Optional[DatasetProfileView] = None,
        column_constraints: Optional[Dict[str, Dict[str, MetricConstraint]]] = None,
    ) -> None:
        if column_constraints is None:
            column_constraints = dict()
        self.column_constraints = column_constraints
        self.dataset_constraints = dict()
        self.dataset_profile_view = dataset_profile_view

    def validate(self, profile_view: Optional[DatasetProfileView] = None) -> bool:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        if len(column_names) == 0:
            logger.warning("validate was called with empty set of constraints, returning True!")
            return True

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                (result, _) = metric_constraint.validate_profile(profile)
                if not result:
                    logger.info(f"{constraint_name} failed on column {column_name}")
                    return False
        metric_names = self.dataset_constraints.keys()
        for metric_name in metric_names:
            metric_constraint = self.dataset_constraints[metric_name]
            (result, _) = metric_constraint.validate_profile(profile)
            if not result:
                logger.info(f"{metric_name} failed on dataset")
                return False
        return True

    @deprecated(message="Please use generate_constraints_report()")
    def report(self, profile_view: Optional[DatasetProfileView] = None) -> List[Tuple[str, int, int]]:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        results: List[Tuple[str, int, int]] = []
        if len(column_names) == 0:
            logger.warning("report was called with empty set of constraints!")
            return results

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                (result, _) = metric_constraint.validate_profile(profile)
                if not result:
                    logger.info(f"{constraint_name} failed on column {column_name}")
                    results.append((constraint_name, 0, 1))
                else:
                    results.append((constraint_name, 1, 0))
        metric_names = self.dataset_constraints.keys()
        for metric_name in metric_names:
            metric_constraint = self.dataset_constraints[metric_name]
            (result, _) = metric_constraint.validate_profile(profile)
            if not result:
                logger.info(f"{metric_name} failed on dataset")
                results.append((metric_name, 0, 1))
            else:
                results.append((metric_name, 1, 0))
        return results

    def _generate_metric_report(
        self,
        profile_view: DatasetProfileView,
        metric_constraint: MetricConstraint,
        constraint_name: str,
        with_summary: bool,
    ) -> ReportResult:
        (result, metric_summary) = metric_constraint.validate_profile(profile_view)
        if not result:
            if with_summary:
                return ReportResult(name=constraint_name, passed=0, failed=1, summary=metric_summary)
            else:
                return ReportResult(name=constraint_name, passed=0, failed=1, summary=None)
        else:
            if with_summary:
                return ReportResult(name=constraint_name, passed=1, failed=0, summary=metric_summary)
            else:
                return ReportResult(name=constraint_name, passed=1, failed=0, summary=None)

    def generate_constraints_report(
        self, profile_view: Optional[DatasetProfileView] = None, with_summary=False
    ) -> List[ReportResult]:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        results: List[ReportResult] = []
        if len(column_names) == 0:
            logger.warning("report was called with empty set of constraints!")
            return results

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                metric_report = self._generate_metric_report(
                    profile_view=profile,
                    metric_constraint=metric_constraint,
                    constraint_name=constraint_name,
                    with_summary=with_summary,
                )
                if metric_report[1] == 0:
                    logger.info(f"{constraint_name} failed on column {column_name}")

                results.append(metric_report)

        metric_names = self.dataset_constraints.keys()
        for metric_name in metric_names:
            metric_constraint = self.dataset_constraints[metric_name]
            metric_report = self._generate_metric_report(
                profile_view=profile,
                metric_constraint=metric_constraint,
                constraint_name=metric_name,
                with_summary=with_summary,
            )
            if metric_report[1] == 0:
                logger.info(f"{metric_name} failed on dataset.")

            results.append(metric_report)

        return results

    def _resolve_profile_view(self, profile_view: Optional[DatasetProfileView]) -> DatasetProfileView:
        if profile_view is None and self.dataset_profile_view is None:
            raise ValueError(
                "Constraints need to be initialized on a profile or have a profile passed in before calling validate or report."
            )
        return profile_view if profile_view is not None else self.dataset_profile_view


class ConstraintsBuilder:
    def __init__(self, dataset_profile_view: DatasetProfileView, constraints: Optional[Constraints] = None) -> None:
        self._dataset_profile_view = dataset_profile_view
        if constraints is None:
            constraints = Constraints(dataset_profile_view=dataset_profile_view)
        self._constraints = constraints

    def get_metric_selectors(self) -> List[MetricsSelector]:
        selectors = []
        column_profiles = self._dataset_profile_view.get_columns()
        for column_name in column_profiles:
            column_profile = column_profiles[column_name]
            metric_component_paths = column_profile._metrics.keys()
            for metric_path in metric_component_paths:
                selectors.append(MetricsSelector(metric_name=metric_path, column_name=column_name))
        return selectors

    def add_constraint(self, constraint: MetricConstraint, ignore_missing: bool = False) -> "ConstraintsBuilder":
        # check that the column exists and we have a column profile view
        column_name = constraint.metric_selector.column_name
        column_profile_view = self._dataset_profile_view.get_column(column_name)
        if column_profile_view is None and not ignore_missing and constraint.metric_selector.metrics_resolver is None:
            raise ValueError(
                f"{column_name} was not found in set of this profile, the list of columns are: {self._dataset_profile_view.get_columns()}"
            )

        metric_selector = constraint.metric_selector
        metrics = metric_selector.apply(self._dataset_profile_view)
        if (metrics is None or len(metrics) == 0) and not (ignore_missing or column_name is None):
            raise ValueError(
                f"metrics not found for column {column_name}, available metric components are: {column_profile_view.get_metric_component_paths()}"
            )

        if column_name is None:
            self._constraints.dataset_constraints[constraint.name] = constraint
        else:
            if column_name not in self._constraints.column_constraints:
                self._constraints.column_constraints[column_name] = dict()
            self._constraints.column_constraints[column_name][constraint.name] = constraint

        return self

    def build(self) -> Constraints:
        return deepcopy(self._constraints)



from .local_store import LocalStore
from .profile_store import ProfileStore
from .query import DatasetIdQuery, DateQuery
from .sqlite_store import SQLiteStore

__ALL__ = [ProfileStore, LocalStore, DateQuery, DatasetIdQuery, SQLiteStore]



import warnings
from typing import Dict, List, Optional, Union

import numpy as np
from scipy import stats  # type: ignore
from scipy.spatial.distance import euclidean
from typing_extensions import TypedDict
from whylogs_sketching import kll_doubles_sketch  # type: ignore

from whylogs.core.utils import deprecated
from whylogs.core.view.column_profile_view import ColumnProfileView  # type: ignore
from whylogs.core.view.dataset_profile_view import DatasetProfileView  # type: ignore
from whylogs.viz.drift.configs import HellingerConfig, KSTestConfig
from whylogs.viz.utils import _calculate_bins
from whylogs.viz.utils.frequent_items_calculations import (
    FrequentStats,
    get_frequent_stats,
    zero_padding_frequent_items,
)


class ColumnDriftValue(TypedDict):
    """p-value for applied statistical test, along with the name of the applied test."""

    p_value: float
    test: str


class ColumnDriftStatistic(TypedDict):
    """Statistic for applied algorithm, along with the name of the applied algorithm."""

    statistic: float
    algorithm: str


def _compute_ks_test_p_value(
    target_distribution: kll_doubles_sketch,
    reference_distribution: kll_doubles_sketch,
    quantiles: Optional[List[float]] = None,
) -> Optional[ColumnDriftValue]:
    """Compute the Kolmogorov-Smirnov test p-value of two continuous distributions.

    Uses the quantile values and the corresponding CDFs to calculate the approximate KS statistic.
    Only applicable to continuous distributions.
    The null hypothesis expects the samples to come from the same distribution.

    Parameters
    ----------
    target_distribution : datasketches.kll_floats_sketch
        A kll_floats_sketch (quantiles sketch) from the target distribution's values
    reference_distribution : datasketches.kll_floats_sketch
        A kll_floats_sketch (quantiles sketch) from the reference (expected) distribution's values
        Can be generated from a theoretical distribution, or another sample for the same feature.
    quantiles: Optional[List[float]], optional
        Bucket of quantiles used to get the CDF's for both target and reference profiles.

    Returns
    -------
        p_value : float
        The estimated p-value from the parametrized KS test, applied on the target and reference distributions'
        kll_floats_sketch summaries

    """

    if not quantiles:
        QUANTILES = KSTestConfig().quantiles
    else:
        QUANTILES = quantiles

    if reference_distribution.is_empty() or target_distribution.is_empty():
        return None

    D_max = 0
    target_quantile_values = target_distribution.get_quantiles(QUANTILES)
    ref_quantile_values = reference_distribution.get_quantiles(QUANTILES)

    num_quantiles = len(QUANTILES)
    i, j = 0, 0
    while i < num_quantiles and j < num_quantiles:
        if target_quantile_values[i] < ref_quantile_values[j]:
            current_quantile = target_quantile_values[i]
            i += 1
        else:
            current_quantile = ref_quantile_values[j]
            j += 1

        cdf_target = target_distribution.get_cdf([current_quantile])[0]
        cdf_ref = reference_distribution.get_cdf([current_quantile])[0]
        D = abs(cdf_target - cdf_ref)
        if D > D_max:
            D_max = D

    while i < num_quantiles:
        cdf_target = target_distribution.get_cdf([target_quantile_values[i]])[0]
        cdf_ref = reference_distribution.get_cdf([target_quantile_values[i]])[0]
        D = abs(cdf_target - cdf_ref)
        if D > D_max:
            D_max = D
        i += 1

    while j < num_quantiles:
        cdf_target = target_distribution.get_cdf([ref_quantile_values[j]])[0]
        cdf_ref = reference_distribution.get_cdf([ref_quantile_values[j]])[0]
        D = abs(cdf_target - cdf_ref)
        if D > D_max:
            D_max = D
        j += 1

    m, n = sorted([target_distribution.get_n(), reference_distribution.get_n()], reverse=True)
    en = m * n / (m + n)

    p_value = stats.distributions.kstwo.sf(D_max, np.round(en))

    return {"p_value": p_value, "test": "ks"}


def _get_ks_p_value(target_view_column, reference_view_column) -> Optional[ColumnDriftValue]:
    target_dist_metric = target_view_column.get_metric("distribution")
    ref_dist_metric = reference_view_column.get_metric("distribution")

    if target_dist_metric is None or ref_dist_metric is None:
        return None

    target_kll_sketch = target_dist_metric.kll.value
    ref_kll_sketch = ref_dist_metric.kll.value

    ks_p_value = _compute_ks_test_p_value(target_kll_sketch, ref_kll_sketch)
    return ks_p_value


def _compute_chi_squared_test_p_value(
    target_distribution: FrequentStats, reference_distribution: FrequentStats
) -> Optional[ColumnDriftValue]:
    """
    Calculate the Chi-Squared test p-value for two discrete distributions.

    Uses the top frequent items summary, unique count estimate and total count estimate for each feature,
    to calculate the estimated Chi-Squared statistic.
    Applicable only to discrete distributions.

    Parameters
    ----------
    target_distribution : ReferenceDistributionDiscreteMessage
        The summary message of the target feature's distribution.
        Should be a ReferenceDistributionDiscreteMessage containing the frequent items,
        unique, and total count summaries.
    reference_distribution : ReferenceDistributionDiscreteMessage
        The summary message of the reference feature's distribution.
        Should be a ReferenceDistributionDiscreteMessage containing the frequent items,
        unique, and total count summaries.

    Returns
    -------
        p_value : ColumnDriftValue or None. ColumnDriftValue has fields `p-value` with the test's result,
        and the name of the test applied in the `test` field (chi-squared).
        The estimated p-value from the Chi-Squared test, applied on the target and reference distributions'
        frequent and unique items summaries
    """
    target_freq_items = target_distribution["frequent_items"]
    ref_freq_items = reference_distribution["frequent_items"]
    target_total_count = target_distribution["total_count"]
    target_unique_count = target_distribution["unique_count"]
    ref_total_count = reference_distribution["total_count"]

    if ref_total_count <= 0 or target_total_count <= 0:
        return None

    target_freq_items, ref_freq_items = zero_padding_frequent_items(target_freq_items, ref_freq_items)

    ref_dist_items = dict()
    for item in reference_distribution["frequent_items"]:
        ref_dist_items[item["value"]] = item["estimate"]
    proportion_ref_dist_items = {k: v / ref_total_count for k, v in ref_dist_items.items()}

    chi_sq = 0.0
    for item in target_freq_items:
        target_frequency = item["estimate"]
        if item["value"] in ref_dist_items:
            ref_frequency = int(proportion_ref_dist_items[item["value"]] * target_total_count)
        else:
            ref_frequency = 0

        if ref_frequency == 0:
            chi_sq = np.inf
            break
        chi_sq += (target_frequency - ref_frequency) ** 2 / ref_frequency

    degrees_of_freedom = target_unique_count - 1
    degrees_of_freedom = degrees_of_freedom if degrees_of_freedom > 0 else 1
    p_value = stats.chi2.sf(chi_sq, degrees_of_freedom)
    return {"p_value": p_value, "test": "chi-squared"}


def _get_chi2_p_value(target_view_column, reference_view_column) -> Optional[ColumnDriftValue]:
    target_frequent_stats: FrequentStats = get_frequent_stats(target_view_column, config=None)
    ref_frequent_stats: FrequentStats = get_frequent_stats(reference_view_column, config=None)

    if not target_frequent_stats or not ref_frequent_stats:
        return None

    chi2_p_value = _compute_chi_squared_test_p_value(
        target_distribution=target_frequent_stats, reference_distribution=ref_frequent_stats
    )
    return chi2_p_value


def calculate_hellinger_distance(target_pmf: List[float], reference_pmf: List[float]) -> float:
    """Calculates hellinger distance between two discrete probability distributions.

    Parameters
    ----------
    target_pmf : List[float]
        Target discrete probability distribution.
    reference_pmf : List[float]
        Reference discrete probability distribution.

    Returns
    -------
    float
        The hellinger distance between the two discrete probability distributions.
        Between 0 (identical distributions) and 1 (completely different distributions).
    """
    # https://en.wikipedia.org/wiki/Hellinger_distance
    distance = euclidean(np.sqrt(target_pmf), np.sqrt(reference_pmf)) / np.sqrt(2)
    return distance


def _get_hellinger_distance(
    target_view_column: ColumnProfileView,
    reference_view_column: ColumnProfileView,
    nbins: Optional[int] = None,
    config: Optional[HellingerConfig] = None,
) -> Optional[ColumnDriftStatistic]:
    if config is None:
        config = HellingerConfig()
    MAX_HIST_BUCKETS = config.max_hist_buckets
    HIST_AVG_NUMBER_PER_BUCKET = config.hist_avg_number_per_bucket
    MIN_N_BUCKETS = config.min_n_buckets
    if MIN_N_BUCKETS < 2:
        warnings.warn(
            "MIN_N_BUCKETS < 2 might lead to erroneous results for low-sized samples. Consider setting it to >=2."
        )

    if not nbins:
        nbins = MAX_HIST_BUCKETS
    target_dist_metric = target_view_column.get_metric("distribution")
    ref_dist_metric = reference_view_column.get_metric("distribution")

    if target_dist_metric is None or ref_dist_metric is None:
        warnings.warn("Column must have a Distribution Metric assigned to it.")
        return None

    target_kll_sketch = target_dist_metric.kll.value
    ref_kll_sketch = ref_dist_metric.kll.value

    if target_kll_sketch.is_empty() or ref_kll_sketch.is_empty():
        warnings.warn("Distribution sketch must not be empty.")
        return None

    start = min([target_kll_sketch.get_min_value(), ref_kll_sketch.get_min_value()])
    end = max([target_kll_sketch.get_max_value(), ref_kll_sketch.get_max_value()])
    n = target_kll_sketch.get_n() + ref_kll_sketch.get_n()
    bins, end = _calculate_bins(
        end=end,
        start=start,
        n=n,
        avg_per_bucket=HIST_AVG_NUMBER_PER_BUCKET,
        max_buckets=nbins,
        min_n_buckets=MIN_N_BUCKETS,
    )

    target_pmf = target_kll_sketch.get_pmf(bins)
    ref_pmf = ref_kll_sketch.get_pmf(bins)
    distance = calculate_hellinger_distance(target_pmf=target_pmf, reference_pmf=ref_pmf)
    return {"statistic": distance, "algorithm": "hellinger"}


@deprecated(message="please use whylogs drift's calculate_drift_score instead")
def calculate_drift_values(
    target_view: DatasetProfileView, reference_view: DatasetProfileView, statistic=False
) -> Dict[str, Optional[Union[ColumnDriftValue, ColumnDriftStatistic]]]:
    """Calculate drift values between both profiles. Applicable for numerical and categorical features.

    Calculates drift only for features found in both profiles, and ignore those not found in either profile.

    Parameters
    ----------
    target_view : DatasetProfileView
        Target Profile View
    reference_view : DatasetProfileView
        Reference Profile View
    statistic: bool
        If false, value will be a pvalue. If true value will be a statistic.


    Returns
    -------
    drift_values: Dict[str, Optional[ColumnDriftValue]]
        A dictionary of the p-values, along with the type of test applied, for the given features.
    """
    drift_values: Dict[str, Optional[Union[ColumnDriftValue, ColumnDriftStatistic]]] = {}
    target_view_columns = target_view.get_columns()
    reference_view_columns = reference_view.get_columns()
    for target_column_name in target_view_columns:
        if target_column_name in reference_view_columns:
            target_view_column = target_view_columns[target_column_name]
            reference_view_column = reference_view_columns[target_column_name]

            if not statistic:
                drift_values[target_column_name] = _get_ks_p_value(
                    target_view_column=target_view_column, reference_view_column=reference_view_column
                ) or _get_chi2_p_value(
                    target_view_column=target_view_column, reference_view_column=reference_view_column
                )
            else:
                drift_values[target_column_name] = _get_hellinger_distance(
                    target_view_column=target_view_column, reference_view_column=reference_view_column
                )
    return drift_values



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['whylogs',
 'whylogs.api',
 'whylogs.api.fugue',
 'whylogs.api.logger',
 'whylogs.api.pyspark.experimental',
 'whylogs.api.reader',
 'whylogs.api.store',
 'whylogs.api.usage_stats',
 'whylogs.api.writer',
 'whylogs.core',
 'whylogs.core.constraints',
 'whylogs.core.constraints.factories',
 'whylogs.core.metrics',
 'whylogs.core.model_performance_metrics',
 'whylogs.core.proto',
 'whylogs.core.proto.v0',
 'whylogs.core.utils',
 'whylogs.core.validators',
 'whylogs.core.view',
 'whylogs.datasets',
 'whylogs.datasets.descr',
 'whylogs.experimental.core.metrics',
 'whylogs.extras',
 'whylogs.migration',
 'whylogs.viz',
 'whylogs.viz.drift',
 'whylogs.viz.enums',
 'whylogs.viz.extensions',
 'whylogs.viz.extensions.reports',
 'whylogs.viz.utils']

package_data = \
{'': ['*'],
 'whylogs.viz': ['html/*',
                 'html/css/*',
                 'html/fonts/*',
                 'html/images/*',
                 'html/js/*',
                 'html/templates/*']}

install_requires = \
['protobuf>=3.19.4', 'whylogs-sketching>=3.4.1.dev3']

extras_require = \
{':extra == "viz" or extra == "whylabs"': ['requests>=2.27,<3.0'],
 ':python_version < "3.11"': ['typing-extensions>=3.10'],
 ':python_version < "3.8"': ['importlib-metadata<4.3'],
 'datasets': ['pandas'],
 'docs': ['sphinx',
          'sphinx-autoapi',
          'sphinx-copybutton>=0.5.0,<0.6.0',
          'myst-parser[sphinx]>=0.17.2,<0.18.0',
          'furo>=2022.3.4,<2023.0.0',
          'sphinx-autobuild>=2021.3.14,<2022.0.0',
          'sphinxext-opengraph>=0.6.3,<0.7.0',
          'sphinx-inline-tabs',
          'ipython_genutils>=0.2.0,<0.3.0',
          'nbsphinx>=0.8.9,<0.9.0',
          'nbconvert>=7.0.0,<8.0.0'],
 'embeddings': ['scikit-learn>=1.0.2,<2.0.0'],
 'fugue': ['fugue>=0.8.0,<0.9.0'],
 'gcs': ['google-cloud-storage>=2.5.0,<3.0.0'],
 'image': ['Pillow>=9.2.0,<10.0.0'],
 'mlflow': ['mlflow-skinny>=1.26.1,<2.0.0'],
 's3': ['boto3>=1.22.13,<2.0.0'],
 'spark': ['pyarrow>=8.0.0,<9.0.0', 'pyspark>=3.0.0,<4.0.0'],
 'viz': ['pybars3>=0.9,<0.10',
         'ipython',
         'numpy',
         'whylabs-client>=0.4.2',
         'Pillow>=9.2.0,<10.0.0'],
 'viz:python_version < "3.11"': ['scipy>=1.5'],
 'whylabs': ['whylabs-client>=0.4.2']}

setup_kwargs = {
    'name': 'whylogs',
    'version': '1.1.23',
    'description': 'Profile and monitor your ML data pipeline end-to-end',
    'long_description': '<img src="https://static.scarf.sh/a.png?x-pxid=bc3c57b0-9a65-49fe-b8ea-f711c4d35b82" /><p align="center">\n<img src="https://i.imgur.com/nv33goV.png" width="35%"/>\n</br>\n\n<h1 align="center">The open standard for data logging\n\n </h1>\n  <h3 align="center">\n   <a href="https://whylogs.readthedocs.io/"><b>Documentation</b></a> &bull;\n   <a href="https://bit.ly/whylogsslack"><b>Slack Community</b></a> &bull;\n   <a href="https://github.com/whylabs/whylogs#python-quickstart"><b>Python Quickstart</b></a> &bull;\n   <a href="https://whylogs.readthedocs.io/en/latest/examples/integrations/writers/Writing_to_WhyLabs.html"><b>WhyLabs Quickstart</b></a>\n </h3>\n\n<p align="center">\n<a href="https://github.com/whylabs/whylogs-python/blob/mainline/LICENSE" target="_blank">\n    <img src="http://img.shields.io/:license-Apache%202-blue.svg" alt="License">\n</a>\n<a href="https://badge.fury.io/py/whylogs" target="_blank">\n    <img src="https://badge.fury.io/py/whylogs.svg" alt="PyPi Version">\n</a>\n<a href="https://github.com/python/black" target="_blank">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Code style: black">\n</a>\n<a href="https://pepy.tech/project/whylogs" target="_blank">\n    <img src="https://pepy.tech/badge/whylogs" alt="PyPi Downloads">\n</a>\n<a href="bit.ly/whylogs" target="_blank">\n    <img src="https://github.com/whylabs/whylogs-python/workflows/whylogs%20CI/badge.svg" alt="CI">\n</a>\n<a href="https://codeclimate.com/github/whylabs/whylogs-python/maintainability" target="_blank">\n    <img src="https://api.codeclimate.com/v1/badges/442f6ca3dca1e583a488/maintainability" alt="Maintainability">\n</a>\n</p>\n\n## What is whylogs\n\nwhylogs is an open source library for logging any kind of data. With whylogs, users are able to generate summaries of their datasets (called _whylogs profiles_) which they can use to:\n\n1. Track changes in their dataset\n2. Create _data constraints_ to know whether their data looks the way it should\n3. Quickly visualize key summary statistics about their datasets\n\nThese three functionalities enable a variety of use cases for data scientists, machine learning engineers, and data engineers:\n\n- Detect data drift in model input features\n- Detect training-serving skew, concept drift, and model performance degradation\n- Validate data quality in model inputs or in a data pipeline\n- Perform exploratory data analysis of massive datasets\n- Track data distributions & data quality for ML experiments\n- Enable data auditing and governance across the organization\n- Standardize data documentation practices across the organization\n- And more\n\n<a href="https://hub.whylabsapp.com/signup" target="_blank">\n    <img src="https://user-images.githubusercontent.com/7946482/193939278-66a36574-2f2c-482a-9811-ad4479f0aafe.png" alt="WhyLabs Signup">\n</a>\n\nIf you have any questions, comments, or just want to hang out with us, please join [our Slack Community](https://bit.ly/rsqrd-slack). In addition to joining the Slack Community, you can also help this project by giving us a ⭐ in the upper right corner of this page.\n\n## Python Quickstart<a name="python-quickstart" />\n\nInstalling whylogs using the pip package manager is as easy as running `pip install whylogs` in your terminal.\n\nFrom here, you can quickly log a dataset:\n\n```python\nimport whylogs as why\nimport pandas as pd\n\n#dataframe\ndf = pd.read_csv("path/to/file.csv")\nresults = why.log(df)\n```\n\nAnd voilà, you now have a whylogs profile. To learn more about about a whylogs profile is and what you can do with it, read on.\n\n## Table of Contents\n\n- [whylogs Profiles](#whylogs-profiles)\n- [Data Constraints](#data-constraints)\n- [Profile Visualization](#profile-visualization)\n- [Integrations](#integrations)\n- [Supported Data Types](#data-types)\n- [Examples](#examples)\n- [Usage Statistics](#usage-statistics)\n- [Community](#community)\n- [Contribute](#contribute)\n\n## whylogs Profiles<a name="whylogs-profiles" />\n\n### What are profiles\n\nwhylogs profiles are the core of the whylogs library. They capture key statistical properties of data, such as the distribution (far beyond simple mean, median, and standard deviation measures), the number of missing values, and a wide range of configurable custom metrics. By capturing these summary statistics, we are able to accurately represent the data and enable all of the use cases described in the introduction.\n\nwhylogs profiles have three properties that make them ideal for data logging: they are **efficient**, **customizable**, and **mergeable**.\n\n<br />\n\n<img align="left" src="https://user-images.githubusercontent.com/7946482/171064257-26bf727e-3480-4ec3-9c9d-5d8a79567bca.png">\n\n**Efficient**: whylogs profiles efficiently describe the dataset that they represent. This high fidelity representation of datasets is what enables whylogs profiles to be effective snapshots of the data. They are better at capturing the characteristics of a dataset than a sample would be—as discussed in our [Data Logging: Sampling versus Profiling](https://whylabs.ai/blog/posts/data-logging-sampling-versus-profiling) blog post—and are very compact.\n\n<br />\n\n<img align="left" src="https://user-images.githubusercontent.com/7946482/171064575-72ee0f76-7365-4fd1-9cab-4debb673baa8.png">\n\n**Customizable**: The statistics that whylogs profiles collect are easily configured and customizable. This is useful because different data types and use cases require different metrics, and whylogs users need to be able to easily define custom trackers for those metrics. It’s the customizability of whylogs that enables our text, image, and other complex data trackers.\n\n<br />\n\n<img align="left" src="https://user-images.githubusercontent.com/7946482/171064525-2d314534-6cdb-4c07-9d9f-5c74d5c03029.png">\n\n**Mergeable**: One of the most powerful features of whylogs profiles is their mergeability. Mergeability means that whylogs profiles can be combined together to form new profiles which represent the aggregate of their constituent profiles. This enables logging for distributed and streaming systems, and allows users to view aggregated data across any time granularity.\n\n<br />\n\n### How do you generate profiles\n\nOnce whylogs is installed, it\'s easy to generate profiles in both Python and Java environments.\n\nTo generate a profile from a Pandas dataframe in Python, simply run:\n\n```python\nimport whylogs as why\nimport pandas as pd\n\n#dataframe\ndf = pd.read_csv("path/to/file.csv")\nresults = why.log(df)\n```\n\n<!---\nFor images, replace `df` with `image="path/to/image.png"`. Similarly, you can profile Python dicts by replacing the dataframe within the `log()` function with a Python `dict` object.\n--->\n\n### What can you do with profiles\n\nOnce you’ve generated whylogs profiles, a few things can be done with them:\n\nIn your local Python environment, you can set data constraints or visualize your profiles. Setting data constraints on your profiles allows you to get notified when your data don’t match your expectations, allowing you to do data unit testing and some baseline data monitoring. With the Profile Visualizer, you can visually explore your data, allowing you to understand it and ensure that your ML models are ready for production.\n\nIn addition, you can send whylogs profiles to the SaaS ML monitoring and AI observability platform [WhyLabs](https://whylabs.ai). With WhyLabs, you can automatically set up monitoring for your machine learning models, getting notified on both data quality and data change issues (such as data drift). If you’re interested in trying out WhyLabs, check out the always free [Starter edition](https://whylabs.ai/free), which allows you to experience the entire platform’s capabilities with no credit card required.\n\n## WhyLabs<a name="whylabs" />\n\nWhyLabs is a managed service offering built for helping users make the most of their whylogs profiles. With WhyLabs, users can ingest profiles and set up automated monitoring as well as gain full observability into their data and ML systems. With WhyLabs, users can ensure the reliability of their data and models, and debug any problems that arise with them.\n\nIngesting whylogs profiles into WhyLabs is easy. After obtaining your access credentials from the platform, you’ll need to set them in your Python environment, log a dataset, and write it to WhyLabs, like so:\n\n```python\nimport whylogs as why\nimport os\n\nos.environ["WHYLABS_DEFAULT_ORG_ID"] = "org-0" # ORG-ID is case-sensitive\nos.environ["WHYLABS_API_KEY"] = "YOUR-API-KEY"\nos.environ["WHYLABS_DEFAULT_DATASET_ID"] = "model-0" # The selected model project "MODEL-NAME" is "model-0"\n\nresults = why.log(df)\n\nresults.writer("whylabs").write()\n```\n\n![image](<https://github.com/whylabs/whylogs/raw/assets/images/chrome-capture-2022-9-4%20(1).gif>)\n\nIf you’re interested in trying out WhyLabs, check out the always free [Starter edition](https://hub.whylabsapp.com/signup), which allows you to experience the entire platform’s capabilities with no credit card required.\n\n## Data Constraints<a name="data-constraints" />\n\nConstraints are a powerful feature built on top of whylogs profiles that enable you to quickly and easily validate that your data looks the way that it should. There are numerous types of constraints that you can set on your data (that numerical data will always fall within a certain range, that text data will always be in a JSON format, etc) and, if your dataset fails to satisfy a constraint, you can fail your unit tests or your CI/CD pipeline.\n\nA simple example of setting and testing a constraint is:\n\n```python\nimport whylogs as why\nfrom whylogs.core.constraints import Constraints, ConstraintsBuilder\nfrom whylogs.core.constraints.factories import greater_than_number\n\nprofile_view = why.log(df).view()\nbuilder = ConstraintsBuilder(profile_view)\nbuilder.add_constraint(greater_than_number(column_name="col_name", number=0.15))\n\nconstraints = builder.build()\nconstraints.report()\n```\n\nTo learn more about constraints, check out: the [Constraints Example](https://bit.ly/whylogsconstraintsexample).\n\n## Profile Visualization<a name="profile-visualization" />\n\nIn addition to being able to automatically get notified about potential issues in data, it’s also useful to be able to inspect your data manually. With the profile visualizer, you can generate interactive reports about your profiles (either a single profile or comparing profiles against each other) directly in your Jupyter notebook environment. This enables exploratory data analysis, data drift detection, and data observability.\n\nTo access the profile visualizer, install the `[viz]` module of whylogs by running `pip install "whylogs[viz]"` in your terminal. One type of profile visualization that we can create is a drift report; here\'s a simple example of how to analyze the drift between two profiles:\n\n```python\nimport whylogs as why\n\nfrom whylogs.viz import NotebookProfileVisualizer\n\nresult = why.log(pandas=df_target)\nprof_view = result.view()\n\nresult_ref = why.log(pandas=df_reference)\nprof_view_ref = result_ref.view()\n\nvisualization = NotebookProfileVisualizer()\nvisualization.set_profiles(target_profile_view=prof_view, reference_profile_view=prof_view_ref)\n\nvisualization.summary_drift_report()\n```\n\n![image](https://user-images.githubusercontent.com/7946482/169669536-a25cce95-acde-4637-b7b9-c2a685f0bc3f.png)\n\nTo learn more about visualizing your profiles, check out: the [Visualizer Example](https://bit.ly/whylogsvisualizerexample)\n\n## Data Types<a name="data-types" />\n\nwhylogs supports both structured and unstructured data, specifically:\n\n| Data type        | Features | Notebook Example                                                                                                                                                |\n| ---------------- | -------- | --------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| Tabular Data     | ✅       | [Getting started with structured data](https://github.com/whylabs/whylogs/blob/mainline/python/examples/basic/Getting_Started.ipynb)                            |\n| Image Data       | ✅       | [Getting started with images](https://github.com/whylabs/whylogs/blob/maintenance/0.7.x/examples/logging_images.ipynb)                                          |\n| Text Data        | ✅       | [String Features](https://github.com/whylabs/whylogs/blob/maintenance/0.7.x/examples/String_Features.ipynb)                                                     |\n| Embeddings       | 🛠        |                                                                                                                                                                 |\n| Other Data Types | ✋       | Do you have a request for a data type that you don’t see listed here? Raise an issue or join our Slack community and make a request! We’re always happy to help |\n\n## Integrations\n\n![current integration](https://user-images.githubusercontent.com/7946482/171062942-01c420f2-7768-4b7c-88b5-e3f291e1b7d8.png)\n\nwhylogs can seamslessly interact with different tooling along your Data and ML pipelines. We have currently built integrations with:\n\n- AWS S3\n- Apache Airflow\n- Apache Spark\n- Mlflow\n- GCS\n\nand much more!\n\nIf you want to check out our complete list, please refer to our [integrations examples](https://github.com/whylabs/whylogs/tree/mainline/python/examples/integrations) page.\n\n## Examples\n\nFor a full set of our examples, please check out the [examples folder](https://github.com/whylabs/whylogs/tree/mainline/python/examples).\n\n## Benchmarks of whylogs\n\nBy design, whylogs run directly in the data pipeline or in a sidecar container and use highly scalable streaming algorithms to compute statistics. Since data logging with whylogs happens in the same infrastructure where the raw data is being processed, it\'s important to think about the compute overhead. For the majority of use cases, the overhead is minimal, usually under 1%. For very large data volumes with thousands of features and 10M+ QPS it can add ~5% overhead. However, for large data volumes, customers are typically in a distributed environment, such as Ray or Apache Spark. This means they benefit from whylogs parallelization—and the map-reducible property of the whylogs profiles keeping the compute overhead to a minimum.\nBelow are benchmarks to demonstrate how efficient whylogs is at processing tabular data with default configurations (tracking distributions, missing values, counts, cardinality, and schema). Two important advantages of this approach are that parallelization speeds up the calculation and whylogs scales with the number of features, rather than the number of rows. Learn more about how whylogs scales here.\n\n| DATA VOLUME                    |       TOTAL COST OF RUNNING WHYLOGS        |                        INSTANCE TYPE                         | CLUSTER SIZE |                                                  PROCESSING TIME |\n| ------------------------------ | :----------------------------------------: | :----------------------------------------------------------: | -----------: | ---------------------------------------------------------------: |\n| 10 GB ~34M rows x 43 columns   |    ~ $ 0.026 per 10 GB, or $2.45 per TB    | c5a.2xlarge, 8 CPU 16GB RAM, $0.308 on demand price per hour |  2 instances | 2.6 minutes of profiling time per instance (running in parallel) |\n| 10 GB, ~34M rows x 43 columns  | ~ $0.016 per 10 GB, estimated $1.60 per TB | c6g.2xlarge, 8 CPU 16GB RAM, $0.272 on demand price per hour |  2 instances | 1.7 minutes of profiling time per instance (running in parallel) |\n| 10 GB ~34M rows x 43 columns   |            ~ $ 0.045 per 10 GB             | c5a.2xlarge, 8 CPU 16GB RAM, $0.308 on demand price per hour | 16 instances |  33 seconds of profiling time per instance (running in parallel) |\n| 80 GB, 83M rows x 119 columns  |             ~ $0.139 per 80 GB             | c5a.2xlarge, 8 CPU 16GB RAM, $0.308 on demand price per hour | 16 instances | 1.7 minutes of profiling time per instance (running in parallel) |\n| 100 GB, 290M rows x 43 columns |            ~ $0.221 per 100 GB             | c5a.2xlarge, 8 CPU 16GB RAM, $0.308 on demand price per hour | 16 instances | 2.7 minutes of profiling time per instance (running in parallel) |\n\n## Usage Statistics<a name="whylogs-profiles" />\n\nStarting with whylogs v1.0.0, whylogs by default collects anonymous information about a user’s environment. These usage statistics do not include any information about the user or the data that they are profiling, only the environment that the user in which the user is running whylogs.\n\nTo read more about what usage statistics whylogs collects, check out the relevant [documentation](https://docs.whylabs.ai/docs/usage-statistics/).\n\nTo turn off Usage Statistics, simply set the `WHYLOGS_NO_ANALYTICS` environment variable to True, like so:\n\n```python\nimport os\nos.environ[\'WHYLOGS_NO_ANALYTICS\']=\'True\'\n```\n\n## Community\n\nIf you have any questions, comments, or just want to hang out with us, please join [our Slack channel](http://join.slack.whylabs.ai/).\n\n## Contribute\n\n### How to Contribute\n\nWe welcome contributions to whylogs. Please see our [contribution guide](https://github.com/whylabs/whylogs/blob/mainline/.github/CONTRIBUTING.md) and our [development guide](https://github.com/whylabs/whylogs/blob/mainline/.github/DEVELOPMENT.md) for details.\n\n### Contributors\n\n<a href="https://github.com/whylabs/whylogs/graphs/contributors">\n  <img src="https://contrib.rocks/image?repo=whylabs/whylogs" />\n</a>\n\nMade with [contrib.rocks](https://contrib.rocks).\n',
    'author': 'WhyLabs.ai',
    'author_email': 'support@whylabs.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://docs.whylabs.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4',
}


setup(**setup_kwargs)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.struct_pb2
import sys
import v0_messages_pb2

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class UniqueCountSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ESTIMATE_FIELD_NUMBER: builtins.int
    UPPER_FIELD_NUMBER: builtins.int
    LOWER_FIELD_NUMBER: builtins.int
    estimate: builtins.float
    upper: builtins.float
    lower: builtins.float
    def __init__(
        self,
        *,
        estimate: builtins.float = ...,
        upper: builtins.float = ...,
        lower: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["estimate", b"estimate", "lower", b"lower", "upper", b"upper"]) -> None: ...

global___UniqueCountSummary = UniqueCountSummary

@typing_extensions.final
class FrequentStringsSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class FrequentItem(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        VALUE_FIELD_NUMBER: builtins.int
        ESTIMATE_FIELD_NUMBER: builtins.int
        value: builtins.str
        estimate: builtins.float
        def __init__(
            self,
            *,
            value: builtins.str = ...,
            estimate: builtins.float = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["estimate", b"estimate", "value", b"value"]) -> None: ...

    ITEMS_FIELD_NUMBER: builtins.int
    @property
    def items(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FrequentStringsSummary.FrequentItem]: ...
    def __init__(
        self,
        *,
        items: collections.abc.Iterable[global___FrequentStringsSummary.FrequentItem] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["items", b"items"]) -> None: ...

global___FrequentStringsSummary = FrequentStringsSummary

@typing_extensions.final
class FrequentNumbersSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class FrequentDoubleItem(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ESTIMATE_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        RANK_FIELD_NUMBER: builtins.int
        estimate: builtins.int
        value: builtins.float
        rank: builtins.int
        def __init__(
            self,
            *,
            estimate: builtins.int = ...,
            value: builtins.float = ...,
            rank: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["estimate", b"estimate", "rank", b"rank", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class FrequentLongItem(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ESTIMATE_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        RANK_FIELD_NUMBER: builtins.int
        estimate: builtins.int
        value: builtins.int
        rank: builtins.int
        def __init__(
            self,
            *,
            estimate: builtins.int = ...,
            value: builtins.int = ...,
            rank: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["estimate", b"estimate", "rank", b"rank", "value", b"value"]) -> None: ...

    DOUBLES_FIELD_NUMBER: builtins.int
    LONGS_FIELD_NUMBER: builtins.int
    @property
    def doubles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FrequentNumbersSummary.FrequentDoubleItem]: ...
    @property
    def longs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FrequentNumbersSummary.FrequentLongItem]: ...
    def __init__(
        self,
        *,
        doubles: collections.abc.Iterable[global___FrequentNumbersSummary.FrequentDoubleItem] | None = ...,
        longs: collections.abc.Iterable[global___FrequentNumbersSummary.FrequentLongItem] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["doubles", b"doubles", "longs", b"longs"]) -> None: ...

global___FrequentNumbersSummary = FrequentNumbersSummary

@typing_extensions.final
class FrequentItemsSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class FrequentItem(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ESTIMATE_FIELD_NUMBER: builtins.int
        JSON_VALUE_FIELD_NUMBER: builtins.int
        estimate: builtins.int
        json_value: builtins.str
        def __init__(
            self,
            *,
            estimate: builtins.int = ...,
            json_value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["estimate", b"estimate", "json_value", b"json_value"]) -> None: ...

    ITEMS_FIELD_NUMBER: builtins.int
    @property
    def items(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FrequentItemsSummary.FrequentItem]: ...
    def __init__(
        self,
        *,
        items: collections.abc.Iterable[global___FrequentItemsSummary.FrequentItem] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["items", b"items"]) -> None: ...

global___FrequentItemsSummary = FrequentItemsSummary

@typing_extensions.final
class CharPosSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class CharPosMapEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___NumberSummary: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___NumberSummary | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    CHARACTER_LIST_FIELD_NUMBER: builtins.int
    CHAR_POS_MAP_FIELD_NUMBER: builtins.int
    character_list: builtins.str
    @property
    def char_pos_map(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___NumberSummary]: ...
    def __init__(
        self,
        *,
        character_list: builtins.str = ...,
        char_pos_map: collections.abc.Mapping[builtins.str, global___NumberSummary] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["char_pos_map", b"char_pos_map", "character_list", b"character_list"]) -> None: ...

global___CharPosSummary = CharPosSummary

@typing_extensions.final
class StringsSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    UNIQUE_COUNT_FIELD_NUMBER: builtins.int
    FREQUENT_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    TOKEN_LENGTH_FIELD_NUMBER: builtins.int
    CHAR_POS_TRACKER_FIELD_NUMBER: builtins.int
    @property
    def unique_count(self) -> global___UniqueCountSummary: ...
    @property
    def frequent(self) -> global___FrequentStringsSummary: ...
    @property
    def length(self) -> global___NumberSummary: ...
    @property
    def token_length(self) -> global___NumberSummary: ...
    @property
    def char_pos_tracker(self) -> global___CharPosSummary: ...
    def __init__(
        self,
        *,
        unique_count: global___UniqueCountSummary | None = ...,
        frequent: global___FrequentStringsSummary | None = ...,
        length: global___NumberSummary | None = ...,
        token_length: global___NumberSummary | None = ...,
        char_pos_tracker: global___CharPosSummary | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["char_pos_tracker", b"char_pos_tracker", "frequent", b"frequent", "length", b"length", "token_length", b"token_length", "unique_count", b"unique_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["char_pos_tracker", b"char_pos_tracker", "frequent", b"frequent", "length", b"length", "token_length", b"token_length", "unique_count", b"unique_count"]) -> None: ...

global___StringsSummary = StringsSummary

@typing_extensions.final
class SchemaSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class TypeCountsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.int
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    INFERRED_TYPE_FIELD_NUMBER: builtins.int
    TYPE_COUNTS_FIELD_NUMBER: builtins.int
    @property
    def inferred_type(self) -> v0_messages_pb2.InferredType: ...
    @property
    def type_counts(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.int]: ...
    def __init__(
        self,
        *,
        inferred_type: v0_messages_pb2.InferredType | None = ...,
        type_counts: collections.abc.Mapping[builtins.str, builtins.int] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["inferred_type", b"inferred_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["inferred_type", b"inferred_type", "type_counts", b"type_counts"]) -> None: ...

global___SchemaSummary = SchemaSummary

@typing_extensions.final
class HistogramSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    START_FIELD_NUMBER: builtins.int
    END_FIELD_NUMBER: builtins.int
    WIDTH_FIELD_NUMBER: builtins.int
    COUNTS_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    BINS_FIELD_NUMBER: builtins.int
    N_FIELD_NUMBER: builtins.int
    start: builtins.float
    end: builtins.float
    width: builtins.float
    @property
    def counts(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    max: builtins.float
    min: builtins.float
    @property
    def bins(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    n: builtins.int
    def __init__(
        self,
        *,
        start: builtins.float = ...,
        end: builtins.float = ...,
        width: builtins.float = ...,
        counts: collections.abc.Iterable[builtins.int] | None = ...,
        max: builtins.float = ...,
        min: builtins.float = ...,
        bins: collections.abc.Iterable[builtins.float] | None = ...,
        n: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bins", b"bins", "counts", b"counts", "end", b"end", "max", b"max", "min", b"min", "n", b"n", "start", b"start", "width", b"width"]) -> None: ...

global___HistogramSummary = HistogramSummary

@typing_extensions.final
class QuantileSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    QUANTILES_FIELD_NUMBER: builtins.int
    QUANTILE_VALUES_FIELD_NUMBER: builtins.int
    @property
    def quantiles(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    @property
    def quantile_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        quantiles: collections.abc.Iterable[builtins.float] | None = ...,
        quantile_values: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["quantile_values", b"quantile_values", "quantiles", b"quantiles"]) -> None: ...

global___QuantileSummary = QuantileSummary

@typing_extensions.final
class NumberSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COUNT_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    MEAN_FIELD_NUMBER: builtins.int
    STDDEV_FIELD_NUMBER: builtins.int
    HISTOGRAM_FIELD_NUMBER: builtins.int
    UNIQUE_COUNT_FIELD_NUMBER: builtins.int
    QUANTILES_FIELD_NUMBER: builtins.int
    FREQUENT_NUMBERS_FIELD_NUMBER: builtins.int
    IS_DISCRETE_FIELD_NUMBER: builtins.int
    count: builtins.int
    min: builtins.float
    max: builtins.float
    mean: builtins.float
    stddev: builtins.float
    @property
    def histogram(self) -> global___HistogramSummary: ...
    @property
    def unique_count(self) -> global___UniqueCountSummary: ...
    @property
    def quantiles(self) -> global___QuantileSummary: ...
    @property
    def frequent_numbers(self) -> global___FrequentNumbersSummary: ...
    is_discrete: builtins.bool
    def __init__(
        self,
        *,
        count: builtins.int = ...,
        min: builtins.float = ...,
        max: builtins.float = ...,
        mean: builtins.float = ...,
        stddev: builtins.float = ...,
        histogram: global___HistogramSummary | None = ...,
        unique_count: global___UniqueCountSummary | None = ...,
        quantiles: global___QuantileSummary | None = ...,
        frequent_numbers: global___FrequentNumbersSummary | None = ...,
        is_discrete: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["frequent_numbers", b"frequent_numbers", "histogram", b"histogram", "quantiles", b"quantiles", "unique_count", b"unique_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["count", b"count", "frequent_numbers", b"frequent_numbers", "histogram", b"histogram", "is_discrete", b"is_discrete", "max", b"max", "mean", b"mean", "min", b"min", "quantiles", b"quantiles", "stddev", b"stddev", "unique_count", b"unique_count"]) -> None: ...

global___NumberSummary = NumberSummary

@typing_extensions.final
class ColumnSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COUNTERS_FIELD_NUMBER: builtins.int
    SCHEMA_FIELD_NUMBER: builtins.int
    NUMBER_SUMMARY_FIELD_NUMBER: builtins.int
    STRING_SUMMARY_FIELD_NUMBER: builtins.int
    FREQUENT_ITEMS_FIELD_NUMBER: builtins.int
    UNIQUE_COUNT_FIELD_NUMBER: builtins.int
    @property
    def counters(self) -> v0_messages_pb2.CountersV0: ...
    @property
    def schema(self) -> global___SchemaSummary: ...
    @property
    def number_summary(self) -> global___NumberSummary: ...
    @property
    def string_summary(self) -> global___StringsSummary: ...
    @property
    def frequent_items(self) -> global___FrequentItemsSummary: ...
    @property
    def unique_count(self) -> global___UniqueCountSummary: ...
    def __init__(
        self,
        *,
        counters: v0_messages_pb2.CountersV0 | None = ...,
        schema: global___SchemaSummary | None = ...,
        number_summary: global___NumberSummary | None = ...,
        string_summary: global___StringsSummary | None = ...,
        frequent_items: global___FrequentItemsSummary | None = ...,
        unique_count: global___UniqueCountSummary | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["counters", b"counters", "frequent_items", b"frequent_items", "number_summary", b"number_summary", "schema", b"schema", "string_summary", b"string_summary", "unique_count", b"unique_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["counters", b"counters", "frequent_items", b"frequent_items", "number_summary", b"number_summary", "schema", b"schema", "string_summary", b"string_summary", "unique_count", b"unique_count"]) -> None: ...

global___ColumnSummary = ColumnSummary

@typing_extensions.final
class DatasetSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class ColumnsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___ColumnSummary: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___ColumnSummary | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    PROPERTIES_FIELD_NUMBER: builtins.int
    COLUMNS_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    @property
    def properties(self) -> v0_messages_pb2.DatasetPropertiesV0: ...
    @property
    def columns(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___ColumnSummary]: ...
    @property
    def model(self) -> global___ModelSummary: ...
    def __init__(
        self,
        *,
        properties: v0_messages_pb2.DatasetPropertiesV0 | None = ...,
        columns: collections.abc.Mapping[builtins.str, global___ColumnSummary] | None = ...,
        model: global___ModelSummary | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["model", b"model", "properties", b"properties"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["columns", b"columns", "model", b"model", "properties", b"properties"]) -> None: ...

global___DatasetSummary = DatasetSummary

@typing_extensions.final
class ModelSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METRICS_FIELD_NUMBER: builtins.int
    @property
    def metrics(self) -> global___MetricsSummary: ...
    def __init__(
        self,
        *,
        metrics: global___MetricsSummary | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["metrics", b"metrics"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["metrics", b"metrics"]) -> None: ...

global___ModelSummary = ModelSummary

@typing_extensions.final
class MetricsSummary(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODEL_TYPE_FIELD_NUMBER: builtins.int
    ROC_FPR_TPR_FIELD_NUMBER: builtins.int
    RECALL_PREC_FIELD_NUMBER: builtins.int
    CONFUSION_MATRIX_FIELD_NUMBER: builtins.int
    model_type: v0_messages_pb2.ModelType.ValueType
    @property
    def roc_fpr_tpr(self) -> global___ROCCurve: ...
    @property
    def recall_prec(self) -> global___RecallCurve: ...
    @property
    def confusion_matrix(self) -> global___ConfusionMatrix: ...
    def __init__(
        self,
        *,
        model_type: v0_messages_pb2.ModelType.ValueType = ...,
        roc_fpr_tpr: global___ROCCurve | None = ...,
        recall_prec: global___RecallCurve | None = ...,
        confusion_matrix: global___ConfusionMatrix | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["confusion_matrix", b"confusion_matrix", "recall_prec", b"recall_prec", "roc_fpr_tpr", b"roc_fpr_tpr"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["confusion_matrix", b"confusion_matrix", "model_type", b"model_type", "recall_prec", b"recall_prec", "roc_fpr_tpr", b"roc_fpr_tpr"]) -> None: ...

global___MetricsSummary = MetricsSummary

@typing_extensions.final
class ConfusionMatrix(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LABELS_FIELD_NUMBER: builtins.int
    TARGET_FIELD_FIELD_NUMBER: builtins.int
    PREDICTIONS_FIELD_FIELD_NUMBER: builtins.int
    SCORE_FIELD_FIELD_NUMBER: builtins.int
    COUNTS_FIELD_NUMBER: builtins.int
    @property
    def labels(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    target_field: builtins.str
    predictions_field: builtins.str
    score_field: builtins.str
    @property
    def counts(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.struct_pb2.ListValue]:
        """e.g. [[33, 6], [11, 27]]"""
    def __init__(
        self,
        *,
        labels: collections.abc.Iterable[builtins.str] | None = ...,
        target_field: builtins.str = ...,
        predictions_field: builtins.str = ...,
        score_field: builtins.str = ...,
        counts: collections.abc.Iterable[google.protobuf.struct_pb2.ListValue] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["counts", b"counts", "labels", b"labels", "predictions_field", b"predictions_field", "score_field", b"score_field", "target_field", b"target_field"]) -> None: ...

global___ConfusionMatrix = ConfusionMatrix

@typing_extensions.final
class ROCCurve(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.struct_pb2.ListValue]:
        """e.g.   "values": [ [1, 0.42857],  [1, 0.42857], ... ]"""
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[google.protobuf.struct_pb2.ListValue] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["values", b"values"]) -> None: ...

global___ROCCurve = ROCCurve

@typing_extensions.final
class RecallCurve(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.struct_pb2.ListValue]:
        """e.g.   "values": [ [1, 1],  [1, 1], ... ]"""
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[google.protobuf.struct_pb2.ListValue] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["values", b"values"]) -> None: ...

global___RecallCurve = RecallCurve

@typing_extensions.final
class DatasetSummaries(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class ProfilesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___DatasetSummary: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___DatasetSummary | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    PROFILES_FIELD_NUMBER: builtins.int
    @property
    def profiles(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___DatasetSummary]: ...
    def __init__(
        self,
        *,
        profiles: collections.abc.Mapping[builtins.str, global___DatasetSummary] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["profiles", b"profiles"]) -> None: ...

global___DatasetSummaries = DatasetSummaries



import logging
from dataclasses import dataclass
from typing import Any

logger = logging.getLogger(__name__)

try:
    import pandas as _pd
except ImportError:  # noqa
    _pd = None  # type: ignore

try:
    import numpy as _np
except ImportError:  # noqa
    _np = None  # type: ignore
    if _pd is not None:
        logger.error("Pandas is installed but numpy is not. Your environment is probably broken.")


class _StubClass:
    pass


@dataclass(frozen=True)
class NumpyStub:
    dtype: type = _StubClass
    number: type = _StubClass
    bool_: type = _StubClass
    floating: type = _StubClass
    ndarray: type = _StubClass
    timedelta64: type = _StubClass
    datetime64: type = _StubClass
    unicode_: type = _StubClass
    issubdtype: type = _StubClass


@dataclass(frozen=True)
class PandasStub(object):
    Series: type = _StubClass
    DataFrame: type = _StubClass


def is_not_stub(stubbed_class: Any) -> bool:
    if stubbed_class and stubbed_class is not _StubClass and not isinstance(stubbed_class, (PandasStub, NumpyStub)):
        return True
    return False


if _np is None:
    _np = NumpyStub()

if _pd is None:
    _pd = PandasStub()

np = _np
pd = _pd
